""" Implements instance extraction and decoding for antecedent trees.

This module implements antecedent trees (Fernandes et al., 2014) within a
framework that expresses coreference resolution as predicting latent structures,
while performing learning using a latent structured perceptron with
cost-augmented inference.

Hence, antecedent trees are expressed as as predicting a latent graph.
In particular, let m_1, ..., m_n be all mentions in a document. Let m_0 be a
dummy mention for anaphoricity determination. We predict
the graph with nodes m_0, ..., m_n and with arcs (m_j, m_i) which correspond to
antecedent decisions. In particular, for each j there exists exactly one i < j
such that (m_j, m_i) is in the graph. Such a graph is called aa *substructure*
(for antecedent trees, substructures and structures coincide).

To implement antecedent trees, this module contains a function that defines the
search space for the graphs, and a decoder that computes the best-scoring tree
of antecedent decisions, and the best-scoring tree of antecedent decisions
consistent with the gold annotation (i.e. only having pairs of coreferent
mentions as arcs).

Reference:

    - Eraldo Fernandes, Cicero dos Santos, and Ruy Milidiu. 2014. Latent trees
      for coreference resolution. *Computational Linguistics*, 40(4):801-835.
      http://www.aclweb.org/anthology/J14-4004
"""

from __future__ import division


import array


from cort.coreference import perceptrons


__author__ = 'martscsn'


def extract_substructures(doc):
    """ Extract the search space for the antecedent tree model,

    The mention ranking model consists in computing the optimal antecedent for
    each anaphor. These decisions are represented as edges in a tree of
    anaphor-antecedent decisions. This functions extracts the search space for
    the tree.

    The search space is represented as a nested list of mention pairs. The
    mention pairs are candidate arcs in the graph. The nested list contains
    only one list, since antecedent trees have only one substructure for
    each document.

    The list contains all potential (anaphor, antecedent) pairs in the
    following order: (m_1, m_0), (m_2, m_1), (m_2, m_0), (m_3, m_2), ...,
    where m_j is the jth mention in the document.

    Args:
        doc (CoNLLDocument): The document to extract substructures from.

    Returns:
        (list(list(Mention, Mention))): The nested list of mention pairs
        describing the search space for the substructures.
    """
    substructure = []

    # iterate over mentions
    for i, ana in enumerate(doc.system_mentions):

        # iterate in reversed order over candidate antecedents
        for ante in sorted(doc.system_mentions[:i], reverse=True):
            substructure.append((ana, ante))

    return [substructure]


class AntecedentTreePerceptron(perceptrons.Perceptron):
    """ A perceptron for antecedent trees. """
    def argmax(self, substructure, arc_information):
        """ Decoder for antecedent trees.

        Compute highest-scoring antecedent tree and highest-scoring antecedent
        tree consistent with the gold annotation.

        Args:
            substructure (list((Mention, Mention))): The list of mention pairs
                which define the search space for one substructure. For mention
                ranking, this list contains all potential anaphor-antecedent
                pairs in the following order:
                (m_1, m_0), (m_2, m_1), (m_2, m_0), (m_3, m_2), ...
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.

        Returns:
            A 7-tuple describing the highest-scoring antecedent tree, and the
            highest-scoring antecedent tree consistent with the gold
            annotation. The tuple consists of:

                - **best_arcs** (*list((Mention, Mention))*): the arcs
                  constituting the highest-scoring antecedent tree,
                - **best_labels** (*list(str)*): empty, the antecedent tree
                  approach does not employ any labels,
                - **best_scores** (*list(float)*): the scores of the
                  arcs in the highest-scoring antecedent tree,
                - **best_cons_arcs** (*list((Mention, Mention))*): the arcs
                  constituting the highest-scoring antecedent tree consistent
                  with the gold annotation.
                - **best_cons_labels** (*list(str)*): empty, the antecedent
                  tree approach does not employ any labels
                - **best_cons_scores** (*list(float)*): the scores of the
                  arcs in the highest-scoring antecedent tree consistent with
                  the gold annotation,
                - **is_consistent** (*bool*): whether the highest-scoring
                  antecedent tree is consistent with the gold annotation.
        """
        if not substructure:
            return [], [], [], [], [], [], True

        number_mentions = len(substructure[0][0].document.system_mentions)

        arcs = []
        arcs_scores = []
        coref_arcs = []
        coref_arcs_scores = []

        is_consistent = True

        for ana_index in range(1, number_mentions):

            first_arc = ana_index*(ana_index-1)//2
            last_arc = first_arc + ana_index

            best, max_val, best_cons, max_cons, best_is_consistent = \
                self.find_best_arcs(substructure[first_arc:last_arc],
                                    arc_information)

            arcs.append(best)
            arcs_scores.append(max_val)
            coref_arcs.append(best_cons)
            coref_arcs_scores.append(max_cons)

            is_consistent &= best_is_consistent

        return (
            arcs,
            [],
            arcs_scores,
            coref_arcs,
            [],
            coref_arcs_scores,
            is_consistent
        )



from collections import defaultdict
import unittest

from cort.analysis import data_structures
from cort.core import documents
from cort.core import mentions
from cort.core import spans


__author__ = 'smartschat'


class TestCorefStructures(unittest.TestCase):
    def setUp(self):
        self.complicated_mention_example = """#begin	document	(test2);	part	000
test2	0	0	This    NN   (NP*	-   -   -   -   -   (0)
test2	0	1	is  NN	*   -   -   -   -   -   -
test2	0	2	just    NN   *	-   -   -   -   -   -
test2	0	3	a   NN   *	-   -   -   -   -   (0|(1)
test2	0	4	test    NN   *	-   -   -   -   -   0)
test2	0	5	.   NN   *)	-   -   -   -   -   -

test2	0	0	It  NN   (NP*	-   -   -   -   -   (1)|(0
test2	0	1	shows   NN   *	-   -   -   -   -   -
test2	0	2	that    NN   *	-   -   -   -   -   (2)
test2	0	3	the NN   *	-   -   -   -   -   (2|(3
test2	0	4	scorer  NN   *	-   -   -   -   -   2)|0)
test2	0	5	works   NN   *	-   -   -   -   -   3)
test2	0	6	.   NN   *)	-   -   -   -   -   -

#end	document"""

        self.complicated_mention_document = documents.CoNLLDocument(
            self.complicated_mention_example)

    def test_entity_graph_from_mentions(self):
        annotated_mentions = \
            self.complicated_mention_document.annotated_mentions

        first_graph = data_structures.EntityGraph({
            annotated_mentions[4]: [annotated_mentions[2],
                                    annotated_mentions[0]],
            annotated_mentions[2]: [annotated_mentions[0]]
        })

        second_graph = data_structures.EntityGraph({
            annotated_mentions[3]: [annotated_mentions[1]]
        })

        third_graph = data_structures.EntityGraph({
            annotated_mentions[6]: [annotated_mentions[5]]
        })

        self.assertEqual(
            [first_graph, second_graph, third_graph],
            data_structures.EntityGraph.from_mentions(annotated_mentions,
                                                       "annotated_set_id"))

    def test_entity_graph_partition(self):
        annotated_mentions = \
            self.complicated_mention_document.annotated_mentions

        graph = data_structures.EntityGraph({
            annotated_mentions[4]: [annotated_mentions[2],
                                    annotated_mentions[0]],
            annotated_mentions[2]: [annotated_mentions[0]]
        })

        system_output = [
            mentions.Mention(
                self.complicated_mention_document,
                spans.Span(0, 0),
                {"set_id": 0}),
            mentions.Mention(
                self.complicated_mention_document,
                spans.Span(2, 3),
                {"set_id": 1}),
            mentions.Mention(
                self.complicated_mention_document,
                spans.Span(6, 10),
                {"set_id": 0}),
            mentions.Mention(
                self.complicated_mention_document,
                spans.Span(5, 5),
                {"set_id": 0})
        ]

        expected_edges = defaultdict(list)
        expected_edges[annotated_mentions[4]].append(annotated_mentions[0])
        expected = data_structures.EntityGraph(expected_edges)

        self.assertEqual(expected,
                         graph.partition(
                             data_structures.EntityGraph.from_mentions(
                                 system_output, "set_id")))


if __name__ == '__main__':
    unittest.main()


""" Manage spans in documents. """

from cort.core import mixins


__author__ = 'smartschat'


class Span(mixins.ComparableMixin):
    """ Manage and compare spans in documents.

    Attributes:
        begin (int): The begin of the span.
        end (int): The end of the span (inclusive).
    """
    def __init__(self, begin, end):
        """ Initialize a span from a begin and an end position.

        Args:
            begin (int): The begin of the span.
            end (int): The end of the span.
        """
        self.begin = begin
        self.end = end

    def __str__(self):
        return "(" + str(self.begin) + ", " + str(self.end) + ")"

    def __repr__(self):
        return "(" + str(self.begin) + ", " + str(self.end) + ")"

    def __lt__(self, other):
        """ Check whether this span is less than another span.

        (a,b) < (c,d) if and only if a < c or a = c and b < d

        Args:
            other (Span): A span.

        Returns:
            True if this span is less than other, False otherwise.
        """
        if self.begin < other.begin:
            return True
        elif self.begin > other.begin:
            return False
        elif self.end < other.end:
            return True
        else:
            return False

    def embeds(self, other):
        """ Check whether this span embeds another span.

        Args:
            other (Span): A span.

        Returns:
            True if this span embeds other, False otherwise.
        """
        return self.begin <= other.begin and self.end >= other.end

    def __hash__(self):
        return hash((self.begin, self.end))

    @staticmethod
    def parse(span_string):
        """ Parse a string specification of a span to a Span object.

        Valid representations are for example "(1, 2)" or "(1,2)".

        Args:
            span_string (str): A string representation of a span.

        Returns:
            Span: The span corresponding to the string representation.
        """
        without_brackets = span_string.strip()[1:-1]
        splitted_and_stripped = [token.strip() for token
                                 in without_brackets.split(",")]
        return Span(
            int(splitted_and_stripped[0]),
            int(splitted_and_stripped[1]))



""" Functions for extracting and filtering mentions in documents. """

from collections import defaultdict
import re

from cort.core import mentions
from cort.core import spans


__author__ = 'smartschat'


def extract_system_mentions(document, filter_mentions=True):
    """ Extract mentions from parse trees and named entity layers in a document.

    Args:
        document (ConLLDocument): The document from which mentions should be
            extracted.
        filter_mentions (bool): Indicates whether extracted mentions should
            be filtered. If set to True, filters:

                - mentions with the same head (retains one with largest span),
                - mentions whose head is embedded in another mention's head,
                - mentions whose head as POS tag JJ,
                - mentions of namend entity type QUANTITY, CARDINAL, ORDINAL,
                  MONEY or PERCENT,
                - mentions "mm", "hmm", "ahem", "um", "US" and "U.S.",
                - non-pronominal mentions embedded in appositions, and
                - pleonastic "it" and "you" detected via heuristics

    Returns:
        list(Mention): the sorted list of extracted system mentions. Includes a
        "dummy mention".
    """
    system_mentions = [mentions.Mention.from_document(span, document)
                       for span in __extract_system_mention_spans(document)]

    if filter_mentions:
        for post_processor in [
            post_process_same_head_largest_span,
            post_process_embedded_head_largest_span,
            post_process_by_head_pos,
            post_process_by_nam_type,
            post_process_weird,
            post_process_appositions,
            post_process_pleonastic_pronoun
        ]:
            system_mentions = post_processor(system_mentions)

    seen = set()

    # update set id and whether it is the first mention in gold entity
    for mention in system_mentions:
        mention.attributes["set_id"] = None

        annotated_set_id = mention.attributes["annotated_set_id"]

        mention.attributes["first_in_gold_entity"] = \
            annotated_set_id not in seen

        seen.add(annotated_set_id)

    system_mentions = [mentions.Mention.dummy_from_document(document)] \
        + system_mentions

    return system_mentions


def __extract_system_mention_spans(document):
    mention_spans = []
    for i, sentence_span in enumerate(document.sentence_spans):
        sentence_tree = document.parse[i]

        in_sentence_spans = __extract_mention_spans_for_sentence(
            sentence_tree,
            document.ner[sentence_span.begin:sentence_span.end+1])

        mention_spans += [spans.Span(sentence_span.begin + span.begin,
                                     sentence_span.begin + span.end)
                          for span in in_sentence_spans]

    return sorted(mention_spans)


def __extract_mention_spans_for_sentence(sentence_tree, sentence_ner):
    return sorted(list(set(
        [__get_in_tree_span(subtree) for subtree
            in sentence_tree.subtrees(__tree_filter)]
        + __get_span_from_ner(
            [pos[1] for pos in sentence_tree.pos()], sentence_ner)
    )))


def __extract_mention_spans_from_tree(sentence_tree):
    mention_spans = []

    for subtree in sentence_tree.subtrees(__tree_filter):
        mention_spans.append(__get_in_tree_span(subtree))

    return sorted(mention_spans)


def __tree_filter(tree):
    return tree.label() == "NP" or tree.label() == "PRP$"


def __get_span_from_ner(pos, ner):
    i = 0
    spans_from_ner = []
    while i < len(ner):
        current_tag = ner[i]
        if current_tag != "NONE":
            start = i

            while i+1 < len(ner) and ner[i+1] != "NONE" and ner[i] == ner[i+1]:
                i += 1

            if i+1 < len(pos) and pos[i+1] == "POS":
                i += 1

            spans_from_ner.append(spans.Span(start, i))

        i += 1

    return sorted(spans_from_ner)


def __get_in_tree_span(parented_tree):
    start = 0

    current_tree = parented_tree

    while current_tree.parent() is not None:
        for child in current_tree.parent():
            if child == current_tree:
                break
            else:
                start += len(child.leaves())

        current_tree = current_tree.parent()

    end = start + len(parented_tree.leaves()) - 1

    return spans.Span(start, end)


def post_process_by_head_pos(system_mentions):
    """ Removes mentions whose head has the part-of-speech tag JJ.

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    return sorted(
        [mention for mention
            in system_mentions
            if not re.match("^(JJ)",
                            mention.attributes["pos"][
                                mention.attributes["head_index"]])]
    )


def post_process_by_nam_type(system_mentions):
    """ Removes proper name mentions of types QUANTITY, CARDINAL, ORDINAL,
    MONEY and PERCENT.

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    return sorted(
        [mention for mention
            in system_mentions
            if mention.attributes["type"] != "NAM" or
            mention.attributes["ner"][mention.attributes["head_index"]] not in
            ["QUANTITY", "CARDINAL", "ORDINAL", "MONEY", "PERCENT"]]
    )


def post_process_weird(system_mentions):
    """ Removes all mentions which are "mm", "hmm", "ahem", "um", "US" or
    "U.S.".

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    return sorted(
        [mention for mention
         in system_mentions
         if " ".join(mention.attributes["tokens"]).lower() not in
         ["mm", "hmm", "ahem", "um"]
         and " ".join(mention.attributes["tokens"]) != "US"
         and " ".join(mention.attributes["tokens"]) != "U.S."]
    )


def post_process_pleonastic_pronoun(system_mentions):
    """ Removes pleonastic it and you.

    These are detected via the following heuristics:
        - it: appears in 'it _ _ that' or 'it _ _ _ that'
        - you: appears in 'you know'

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    filtered = []

    for mention in system_mentions:
        if " ".join(mention.attributes["tokens"]).lower() == "it":
            context_two = mention.get_context(2)
            context_three = mention.get_context(3)

            if context_two is not None:
                if context_two[-1] == "that":
                    continue

            if context_three is not None:
                if context_three[-1] == "that":
                    continue

        if " ".join(mention.attributes["tokens"]).lower() == "you":
            if mention.get_context(1) == ["know"]:
                continue

        filtered.append(mention)

    return sorted(filtered)


def post_process_same_head_largest_span(system_mentions):
    """ Removes a mention if there exists a larger mention with the same head.

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    head_span_to_mention = defaultdict(list)

    for mention in system_mentions:
        head_span_to_mention[mention.attributes["head_span"]].append(
            (mention.span.end - mention.span.begin, mention))

    return sorted([sorted(head_span_to_mention[head_span])[-1][1]
                   for head_span in head_span_to_mention])


def post_process_embedded_head_largest_span(system_mentions):
    """ Removes a mention its head is embedded in another head.

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    map_for_heads = {}

    for mention in system_mentions:
        head_span = mention.attributes["head_span"]
        if head_span.end not in map_for_heads:
            map_for_heads[head_span.end] = []

        map_for_heads[head_span.end].append(head_span.begin)

    post_processed_mentions = []

    for mention in system_mentions:
        head_span = mention.attributes["head_span"]
        head_begins = sorted(map_for_heads[head_span.end])
        if head_begins[0] < head_span.begin:
            continue
        else:
            post_processed_mentions.append(mention)

    return sorted(post_processed_mentions)


def post_process_appositions(system_mentions):
    """ Removes a mention its embedded in an apposition.

    Args:
        system_mentions (list(Mention): A list of system mentions.

    Returns:
        list(Mention): the filtered list of mentions.
    """
    appos = [mention for mention
             in system_mentions if mention.attributes["is_apposition"]]

    post_processed_mentions = []

    for mention in system_mentions:
        span = mention.span
        embedded_in_appo = False
        for appo in appos:
            appo_span = appo.span
            if appo_span.embeds(span) and appo_span != span:
                if len(appo.attributes["parse_tree"]) == 2:
                    embedded_in_appo = True
                elif (mention.attributes["parse_tree"] in
                        appo.attributes["parse_tree"]):
                    embedded_in_appo = True

        if mention.attributes["type"] == "PRO" or not embedded_in_appo:
            post_processed_mentions.append(mention)

    return sorted(post_processed_mentions)



""" Contains perceptron learning methods for structured prediction. """


from __future__ import division

import logging
import array

import numpy

cimport numpy
cimport cython


__author__ = 'smartschat'


cdef class Perceptron:
    cdef int n_iter, random_seed
    cdef double cost_scaling
    cdef dict priors, weights, label_to_index
    """ Provide a latent structured perceptron.

    This implementation provides a latent structured perceptron with
    cost-augmented inference and parameter averaging for graphs encoding
    coreference decisions.

    Attributes:
        n_iter (int): The number of epochs for training. Defaults to 5.
        seed (int): The random seed for shuffling the data. Defaults to 23.
        cost_scaling (int): The parameter for scaling the cost function during
            cost-augmented inference. Defaults to 1.
        priors (dict(str, float)): A mapping of graph labels to priors for
            these labels.
        weights (dict(str, array)): A mapping of labels to weight
            vectors. For each label ``l``, ``weights[l]`` contains weights
            for each feature seen during training (for representing the
            features we employ *feature hashing*). If the graphs employed are
            not labeled, ``l`` is set to "+".

    """

    def __init__(self,
                 n_iter=5,
                 seed=23,
                 cost_scaling=1,
                 priors=None,
                 weights=None,
                 cluster_features=None):
        """
        Initialize the perceptron.

        Args:
            n_iter (int): The number of epochs for training. Defaults to 5.
            seed (int): The random seed for shuffling the data. Defaults to 23.
            cost_scaling (int): The parameter for scaling the cost function
                during cost-augmented inference. Defaults to 1.
            label (list(str)): A list of labels used in the graphs. If
                ``None``, defaults to ``["+"]``.
            priors (dict(str, float)): A mapping of graph labels to priors
                for these labels. If ``None`` defaults to an empty mapping.
            weights (dict(str, array)): A mapping of graph labels to
                arrays. For each label ``l``, ``weights[l]`` contains
                weights for each feature seen during training (for representing
                the features we employ *feature hashing*). If the graphs
                employed are not labeled, ``l`` is set to "+".
                If ``None`` defaults to a mapping of "+" to an array only
                containing 0s.
        """
        cdef double[:] weights_for_label

        self.n_iter = n_iter
        self.random_seed = seed
        self.cost_scaling = cost_scaling

        labels = self.get_labels()

        self.label_to_index = {}
        for i, label in enumerate(labels):
            self.label_to_index[label] = i

        if not priors:
            self.priors = {}
            for label in labels:
                self.priors[label] = 0.0
        else:
            self.priors = priors

        self.weights = {}

        try:
            if weights is None:
                for label in labels:
                    weights_for_label = array.array("d",
                                                    (0.0 for k in range(2**24)))
                    self.weights[label] = weights_for_label
            else:
                for label in weights:
                    weights_for_label = weights[label]
                    self.weights[label] = weights_for_label
        # in python 2, array.array does not support the buffer interface
        except TypeError:
            if weights is None:
                for label in labels:
                    weights_for_label = numpy.zeros(2**24, dtype=float)
                    self.weights[label] = weights_for_label
            else:
                for label in weights:
                    weights_for_label = numpy.array(weights[label])
                    self.weights[label] = weights_for_label

    def fit(self, substructures, arc_information):
        """Learn weights from data.

        The learned label priors and weights are stored in ``self.priors`` and
        ``self.weights``.

        Args:
            substructures (list(list((Mention, Mention)))): The search space
                for the substructures, defined by a nested list. The ith list
                contains the search space for the ith substructure.
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.
        """

        cdef double[:] cached_weights_for_label

        indices = list(range(0, len(substructures)))
        numpy.random.seed(self.random_seed)

        cached_priors = {}
        cached_weights = {}

        for label in self.priors:
            cached_priors[label] = 0.0

            try:
                cached_weights_for_label = array.array("d",
                                                    (0.0 for k in range(2**24)))
            except TypeError:
                cached_weights_for_label = numpy.zeros(2**24, dtype=float)

            cached_weights[label] = cached_weights_for_label

        counter = 0

        for epoch in range(1, self.n_iter+1):
            logging.info("Started epoch " + str(epoch))
            numpy.random.shuffle(indices)

            incorrect = 0

            for i in indices:
                substructure = substructures[i]

                (arcs,
                 arcs_labels,
                 arcs_scores,
                 cons_arcs,
                 cons_labels,
                 cons_scores,
                 is_consistent) = self.argmax(substructure,
                                              arc_information)

                if not is_consistent:
                    self.__update(cons_arcs,
                                  arcs,
                                  cons_labels,
                                  arcs_labels,
                                  arc_information,
                                  counter,
                                  cached_priors,
                                  cached_weights)

                    incorrect += 1

                counter += 1

            logging.info("Finished epoch " + str(epoch))
            logging.info("\tIncorrect predictions: " + str(incorrect) + "/" +
                         str(len(indices)))

        # averaging
        for label in self.priors:
            self.priors[label] -= (1/counter)*cached_priors[label]
            self._average_weights(self.weights[label], cached_weights[label],
                                  1.0*counter)

    def predict(self, substructures, arc_information):
        """
        Predict coreference information according to a learned model.

        Args:
            substructures (list(list((Mention, Mention)))): The search space
                for the substructures, defined by a nested list. The ith list
                contains the search space for the ith substructure.
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.
        Returns:
            Three nested lists describing the output. In particular, these
            lists are:

                - arcs (list(list(Mention, Mention))): The nested list of
                  predicted arcs. The ith list contains predictions for the
                  ith substructure.
                - labels (list(list(str))): Labels of the predicted arcs.
                - arcs (list(list(float))): Scores for the predicted arcs.
        """
        arcs = []
        labels = []
        scores = []

        for substructure in substructures:
            (substructure_arcs, substructure_arcs_labels,
             substructure_arcs_scores, _, _, _, _) = self.argmax(
                substructure, arc_information)

            arcs.append(substructure_arcs)
            labels.append(substructure_arcs_labels)
            scores.append(substructure_arcs_scores)

        return arcs, labels, scores

    def argmax(self, substructure, arc_information):
        """ Decoder for coreference resolution.

        Compute highest-scoring substructure and highest-scoring constrained
        substructure consistent with the gold annotation. To implement
        coreference resolution approaches, inherit this class and implement
        this function.

        Args:
            substructure (list((Mention, Mention))): The list of mention pairs
                which define the search space for one substructure.
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.

        Returns:
            A 7-tuple describing the highest-scoring substructure and the
            highest-scoring substructure consistent with the gold information.
            The tuple consists of:

                - **best_arcs** (*list((Mention, Mention))*): the list of arcs
                   in the highest-scoring substructure,
                - **best_labels** (*list(str)*): the list of labels of the
                  arcs in the highest-scoring substructure,
                - **best_scores** (*list(float)*): the scores of the arcs in
                  the highest-scoring substructure,
                - **best_cons_arcs** (*list((Mention, Mention))*): the list of
                  arcs in the highest-scoring constrained substructure
                  consistent with the gold information,
                - **best_cons_labels** (*list(str)*): the list of labels of the
                  arcs in the highest-scoring constrained substructure
                  consistent with the gold information,
                - **best_cons_scores** (*list(float)*): the scores of the arcs
                  in the highest-scoring constrained substructure consistent
                  with the gold information,
                - **is_consistent** (*bool*): whether the highest-scoring
                  substructure is consistent with the gold information.
        """
        raise NotImplementedError()

    def __update(self, good_arcs, bad_arcs, good_labels, bad_labels,
                 arc_information, counter, cached_priors, cached_weights):

        if good_labels or bad_labels:
            for arc, label in zip(good_arcs, good_labels):
                nonnumeric_features, numeric_features, numeric_vals = \
                    arc_information[arc][0]

                self._update_cython(self.weights[label],
                               cached_weights[label],
                               nonnumeric_features,
                               numeric_features,
                               numeric_vals,
                               1.0,
                               1.0*counter)

                self.priors[label] += 1
                cached_priors[label] += counter

            for arc, label in zip(bad_arcs, bad_labels):
                nonnumeric_features, numeric_features, numeric_vals = \
                    arc_information[arc][0]

                self._update_cython(self.weights[label],
                               cached_weights[label],
                               nonnumeric_features,
                               numeric_features,
                               numeric_vals,
                               -1.0,
                               -1.0*counter)

                self.priors[label] -= 1
                cached_priors[label] -= counter
        else:
            for arc in good_arcs:
                nonnumeric_features, numeric_features, numeric_vals = \
                    arc_information[arc][0]

                self._update_cython(self.weights["+"],
                               cached_weights["+"],
                               nonnumeric_features,
                               numeric_features,
                               numeric_vals,
                               1.0,
                               1.0*counter)

            for arc in bad_arcs:
                nonnumeric_features, numeric_features, numeric_vals = \
                    arc_information[arc][0]

                self._update_cython(self.weights["+"],
                               cached_weights["+"],
                               nonnumeric_features,
                               numeric_features,
                               numeric_vals,
                               -1.0,
                               -1.0*counter)

    def get_labels(self):
        """ Get the graph labels employed by the current approach.

        In this case, the only graph label employed is '+'.

        Returns:
            list(str): A list of graph labels (contains only one graph label
                '+').
        """
        return ["+"]

    def get_coref_labels(self):
        """ Get the graph labels signalling coreference employed by the
        current approach.

        In this case, the only graph label signalling coreference is '+'.

        Returns:
            list(str): A list of graph labels signalling coreference (contains
                only one graph label '+').
        """
        return ["+"]

    def find_best_arcs(self, arcs, arc_information, label="+"):
        """ Find the highest-scoring arc and arc consistent with the gold
        information among a set of arcs.

        Args:
            arcs (list((Mention, Mention)): A list of mention pairs constituting
                arcs.
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.
            label (str): The label of the arcs. Defaults to "+".

        Returns:
            A 5-tuple describing the highest-scoring anaphor-antecedent
            decision, and the highest-scoring anaphor-antecedent decision
            consistent with the gold annotation. The tuple consists of:

                - **best** (*(Mention, Mention)*): the highest-scoring
                  anaphor-antecedent decision.
                - **max_val** (*float*): the score of the highest-scoring
                  anaphor-antecedent decision,
                - **best_cons** (*(Mention, Mention)*): the highest-scoring
                  anaphor-antecedent decision consistent with the gold
                  annotation.
                - **max_const** (*float*): the score of the highest-scoring
                  anaphor-antecedent decision consistent with the gold
                  annotation.
                - **is_consistent** (*bool*): whether the highest-scoring
                  anaphor-antecedent decision is consistent with the gold
                  information.
        """
        max_val = float("-inf")
        best = None

        max_cons = float("-inf")
        best_cons = None

        best_is_consistent = False

        for arc in arcs:
            features, costs, consistent = arc_information[arc]

            nonnumeric_features, numeric_features, numeric_vals = features

            score = self._cython_score_arc(self.priors[label],
                                      self.weights[label],
                                      self.cost_scaling,
                                      costs[self.label_to_index[label]],
                                      nonnumeric_features,
                                      numeric_features,
                                      numeric_vals)

            if score > max_val:
                best = arc
                max_val = score
                best_is_consistent = consistent

            if score > max_cons and consistent:
                best_cons = arc
                max_cons = score

        return best, max_val, best_cons, max_cons, best_is_consistent

    def score_arc(self, arc, arc_information, label="+"):
        """ Score an arc according to priors, weights and costs.

        Args:
            arc ((Mention, Mention)): The pair of mentions constituting the arc.
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.
            label (str): The label of the arc. Defaults to "+".

        Returns:
            float: The sum of all weights for the features, plus the scaled
                costs for predicting the arc, plus the prior for the label.
        """

        features, costs, consistent = arc_information[arc]

        nonnumeric_features, numeric_features, numeric_vals = features

        return self._cython_score_arc(
            self.priors[label],
            self.weights[label],
            self.cost_scaling,
            costs[self.label_to_index[label]],
            nonnumeric_features,
            numeric_features,
            numeric_vals
        )

    def get_model(self):
        """ Get the priors and weights of the current model.

        Returns:
            A tuple containing priors and weights. The tuple consists of:

            - **priors** (**dict(str, float)**): A mapping of graph labels to
                priors for these labels.
            - **weights** (**dict(str, array)**): A mapping of graph labels to
                arrays. For each label ``l``, ``weights[l]`` contains
                weights for each feature seen during training (for representing
                the features we employ *feature hashing*). If the graphs
                employed are not labeled, ``l`` is set to "+".
        """
        array_weights = {}

        for label in self.weights:
            array_weights[label] = array.array("d", self.weights[label])

        return self.priors, array_weights

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef double _cython_score_arc(self,
                                  double prior,
                                  double[:] weights,
                                  double cost_scaling,
                                  double costs,
                                  numpy.uint32_t[:] nonnumeric_features,
                                  numpy.uint32_t[:] numeric_features,
                                  float[:] numeric_vals):

        cdef double score = 0.0
        cdef int index = 0

        score += prior
        score += cost_scaling * costs

        for index in range(nonnumeric_features.shape[0]):
            score += weights[nonnumeric_features[index]]

        for index in range(numeric_features.shape[0]):
            score += weights[numeric_features[index]]*numeric_vals[index]

        return score

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void _update_cython(self,
                             double[:] weights,
                             double[:] cached_weights,
                             numpy.uint32_t[:] nonnumeric_features,
                             numpy.uint32_t[:] numeric_features,
                             float[:] numeric_vals,
                             double update_val_for_weights,
                             double update_val_for_cached_weights):
        cdef int index

        for index in range(nonnumeric_features.shape[0]):
            weights[nonnumeric_features[index]] += update_val_for_weights
            cached_weights[nonnumeric_features[index]] += \
                update_val_for_cached_weights

        for index in range(numeric_features.shape[0]):
            weights[numeric_features[index]] += \
                update_val_for_weights*numeric_vals[index]
            cached_weights[numeric_features[index]] += \
                update_val_for_cached_weights*numeric_vals[index]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _average_weights(self, double[:] weights, double[:] cached_weights,
                          double counter):
        cdef int index

        for index in range(weights.shape[0]):
            weights[index] -= cached_weights[index]/counter


""" Read in and access data from external resources such as gender lists."""

import os
import pickle


import cort
from cort.core import singletons
from cort.core import util


__author__ = 'smartschat'


@singletons.Singleton
class GenderData:
    """ Read in and access data from lists with gender information.

    Attributes:
        word_to_gender (dict(str, str)): A mapping from lower-case strings
            to one of four genders: 'MALE', 'FEMALE', 'NEUTRAL' and 'PLURAL'.
    """
    def __init__(self):
        """ Initialize the word-to-gender mapping from gender lists.
        """
        self.word_to_gender = {}

        directory = cort.__path__[0] + "/resources/"

        lists = [
            open(directory + "male.list"),
            open(directory + "female.list"),
            open(directory + "neutral.list"),
            open(directory + "plural.list")
        ]

        genders = ["MALE", "FEMALE", "NEUTRAL", "PLURAL"]

        for gender, gender_list in zip(genders, lists):
            for word in gender_list.readlines():
                self.word_to_gender[word.strip()] = gender

    def look_up(self, attributes):
        """ Look up the gender of a mention described by the input attributes.

        Args:
            attributes (dict(str,object)): A dict describing attributes of
                mentions. Must contain "tokens" and "head", which have lists
                of strings as values.

        Returns:
            (str): None or one of the four genders 'MALE', 'FEMALE',
            'NEUTRAL' or 'PLURAL'.
        """
        # whole string
        if " ".join(attributes["tokens"]).lower() in self.word_to_gender:
            return self.word_to_gender[" ".join(attributes["tokens"]).lower()]
        # head
        elif " ".join(attributes["head"]).lower() in self.word_to_gender:
            return self.word_to_gender[" ".join(attributes["head"]).lower()]
        # head token by token
        elif self.__look_up_token_by_token(attributes["head"]):
            return self.__look_up_token_by_token(attributes["head"])

    def __look_up_token_by_token(self, tokens):
        for token in tokens:
            if token[0].isupper() and token.lower() in self.word_to_gender:
                return self.word_to_gender[token.lower()]


@singletons.Singleton
class LexicalData:
    """ Read in and access data containing pairs of coreferent mention strings.

    Attributes:
        pairs (set((str, str))): A set of string pairs, which represent strings
            of potentially coreferent mentions.
    """
    def __init__(self):
        """ Initialize the set of pairs from
            package_root/resources/coreferent_pairs.obj.
        """
        directory = cort.__path__[0] + "/resources/"

        self.pairs = pickle.load(
            open(directory + "coreferent_pairs.obj", "rb"))

    def look_up(self, anaphor, antecedent):
        """ Look up strings of the mentions in the pair list.

        Args:
            anaphor (Mention): A mention.
            antecedent (Mention): Another mention, the candidate antecedent
                for anaphor.

        Returns:
            True if the pair of strings corresponding to anaphor of
            antecedent, stripped determiners and possessive s, can be found
            in the list of pairs.
        """
        # whole string
        anaphor_cleaned = " ".join(
            util.clean_via_pos(anaphor.attributes["tokens"],
                          anaphor.attributes["pos"]))
        antecedent_cleaned = " ".join(
            util.clean_via_pos(antecedent.attributes["tokens"],
                               antecedent.attributes["pos"]))

        return (
            (anaphor_cleaned, antecedent_cleaned) in self.pairs
            or (antecedent_cleaned, anaphor_cleaned) in self.pairs
        )


@singletons.Singleton
class SingletonMentions:
    """ Read in and access data strings of singleton mentions.

    Attributes:
        singletons (set(str)): A set of strings, which represent strings of
            of potential singleton mentions.
    """
    def __init__(self):
        """ Initialize the set of pairs from
            package_root/resources/singletons_not_cleaned.obj.
        """
        directory = cort.__path__[0] + "/resources/"

        self.singletons = pickle.load(
            open(directory + "singletons_not_cleaned.obj", "rb"))



__author__ = 'martscsn'

import cort

import codecs

import stanford_corenlp_pywrapper

from StanfordDependencies import CoNLL

from cort.core import corpora, documents, spans

import bs4


class Pipeline():
    def __init__(self, corenlp_location, with_coref=False):
        package_dir = cort.__path__[0]

        if with_coref:
            self.proc = stanford_corenlp_pywrapper.CoreNLP(
                configfile=package_dir + "/config_files/corenlp_with_coref.ini",
                corenlp_jars=[corenlp_location + "/*"]
            )
        else:
            self.proc = stanford_corenlp_pywrapper.CoreNLP(
                configfile=package_dir + "/config_files/corenlp.ini",
                corenlp_jars=[corenlp_location + "/*"]
            )

        self.with_coref = with_coref

    def run_on_docs(self, identifier, docs):
        processed_documents = []

        for doc in docs:
            processed_documents.append(self.run_on_doc(
                codecs.open(doc, "r", "utf-8")
            ))

        return corpora.Corpus(identifier, processed_documents)

    def run_on_doc(self, doc_file, name=None):
        if self.with_coref:
            soup = bs4.BeautifulSoup(doc_file.read())
            preprocessed = self.proc.parse_doc(soup.text)
        else:
            data = doc_file.read()
            preprocessed = self.proc.parse_doc(data)

        sentences = []

        for sentence in preprocessed["sentences"]:
            processed_ner = []
            for ner in sentence["ner"]:
                if ner == "O" or ner == "MISC":
                    processed_ner.append("NONE")
                else:
                    processed_ner.append(ner)

            processed_dep = []

            index_to_dep_info = {}
            for dep_info in sentence["deps_basic"]:
                label, head, in_sent_index = dep_info
                index_to_dep_info[in_sent_index] = label, head

            for i in range(0, len(sentence["tokens"])):
                if i in index_to_dep_info.keys():
                    label, head = index_to_dep_info[i]
                    processed_dep.append(
                        CoNLL.Token(
                            form=sentence["tokens"][i],
                            lemma=sentence["lemmas"][i],
                            pos=sentence["pos"][i],
                            index=i+1,
                            head=head+1,
                            deprel=label,
                            cpos=None,
                            feats=None,
                            phead=None,
                            pdeprel=None
                        )
                    )
                else:
                    processed_dep.append(
                        CoNLL.Token(
                            form=sentence["tokens"][i],
                            lemma=sentence["lemmas"][i],
                            pos=sentence["pos"][i],
                            index=i+1,
                            head=0,
                            deprel="punc",
                            cpos=None,
                            feats=None,
                            phead=None,
                            pdeprel=None
                        )
                    )

            sentences.append(
                (sentence["tokens"],
                 sentence["pos"],
                 processed_ner,
                 ["-"]*len(sentence["tokens"]),
                 sentence["parse"],
                 processed_dep,
                )
            )

        if not name:
            name = doc_file.name

        if self.with_coref:
            antecedent_decisions = {}
            coref = {}

            mention_id_to_spans = {}

            max_entity = 0

            for mention in soup.findAll("mention"):
                if mention.get("entity"):
                    max_entity = max(max_entity, int(mention.get("entity")))

            for mention in soup.findAll("mention"):
                mention_id = int(mention.get("id"))

                span = spans.Span(int(mention.get("span_start")),
                                  int(mention.get("span_end")))

                mention_id_to_spans[mention_id] = span

                if mention.get("entity"):
                    annotated_set_id = int(mention.get("entity"))
                else:
                    annotated_set_id = max_entity + 1 + mention_id

                coref[span] = annotated_set_id

                if mention.get("antecedent"):
                    antecedent_decisions[span] = mention_id_to_spans[
                        int(mention.get("antecedent"))
                    ]

            doc = documents.Document(
                name,
                sentences,
                coref)

            spans_to_annotated_mentions = {}

            for mention in doc.annotated_mentions:
                spans_to_annotated_mentions[mention.span] = mention

            for span in antecedent_decisions:
                ante_span = antecedent_decisions[span]
                ana = spans_to_annotated_mentions[span]
                ante = spans_to_annotated_mentions[ante_span]
                ana.attributes["antecedent"] = ante
        else:
            doc = documents.Document(
                name,
                sentences,
                {})

        return doc



__author__ = 'smartschat'


def for_each_relation_with_distance(anaphor,
                                    antecedent,
                                    relations,
                                    relation_weights):
    weight = 0.0

    if len(relations["negative_relations"]) > 0:
        return float("-inf")

    if len(relations["positive_relations"]) == 0:
        return 0

    for relation in relations["positive_relations"]:
        weight += relation_weights[relation]

    weight /= (anaphor.attributes["sentence_id"] -
               antecedent.attributes["sentence_id"]
               + 1)

    return weight



import re

from cort.core import external_data
from cort.core import spans
from cort.core import util


__author__ = 'smartschat'


def not_singleton(anaphor, antecedent):
    singleton_data = external_data.SingletonMentions.get_instance()
    anaphor = " ".join(anaphor.attributes["tokens"])
    antecedent = " ".join(antecedent.attributes["tokens"])

    if (anaphor in singleton_data.singletons and
            singleton_data.singletons[anaphor] >= 25):
        return True

    if (antecedent in singleton_data.singletons and
            singleton_data.singletons[antecedent] >= 25):
        return True


def pronoun_parallelism(anaphor, antecedent):
    return (anaphor.attributes["type"] == "PRO"
            and (anaphor.attributes["citation_form"]
                 in ["he", "she", "it", "they"])
            and (antecedent.attributes["type"] != "PRO"
                 or (antecedent.attributes["citation_form"]
                     in ["he", "she", "it", "they"]))
            and (antecedent.attributes["grammatical_function"] ==
                 anaphor.attributes["grammatical_function"])
            and (antecedent.attributes["grammatical_function"]
                 in ["SUBJECT", "OBJECT"]))


def antecedent_is_subject(anaphor, antecedent):
    return (anaphor.attributes["type"] == "PRO"
            and (anaphor.attributes["citation_form"]
                 in ["he", "she", "it", "they"])
            and (antecedent.attributes["type"] != "PRO"
                 or (antecedent.attributes["citation_form"]
                     in ["he", "she", "it", "they"]))
            and antecedent.attributes["grammatical_function"] == "SUBJECT")


def antecedent_is_object(anaphor, antecedent):
    return (anaphor.attributes["type"] == "PRO"
            and (anaphor.attributes["citation_form"]
                 in ["he", "she", "it", "they"])
            and (antecedent.attributes["type"] != "PRO"
                 or (antecedent.attributes["citation_form"]
                     in ["he", "she", "it", "they"]))
            and antecedent.attributes["grammatical_function"] == "OBJECT")


def anaphor_pronoun(anaphor, antecedent):
    return (anaphor.attributes["type"] == "PRO"
            and (anaphor.attributes["citation_form"]
                 in ["he", "she", "it", "they"])
            and (antecedent.attributes["type"] != "PRO"
                 or (antecedent.attributes["citation_form"]
                     in ["he", "she", "it", "they"])))


def lexical(anaphor, antecedent):
    lexical_data = external_data.LexicalData.get_instance()
    if ((anaphor.attributes["type"] == "NAM"
         and antecedent.attributes["type"] == "NAM")
        or (anaphor.attributes["type"] == "NOM"
            and anaphor.attributes["fine_type"] == "DEF"
            and antecedent.attributes["type"] in ["NAM", "NOM"])):
        return lexical_data.look_up(anaphor, antecedent)


def non_pronominal_string_match(anaphor, antecedent):
    if anaphor.attributes["type"] in ["PRO", "DEM", "VRB"]:
        return False
    elif antecedent.attributes["type"] in ["PRO", "DEM", "VRB"]:
        return False
    else:
        return (" ".join(util.clean_via_pos(anaphor.attributes["tokens"],
                                            anaphor.attributes["pos"])).lower()
                == " ".join(util.clean_via_pos(
                            antecedent.attributes["tokens"],
                            antecedent.attributes["pos"])).lower())


def head_match(anaphor, antecedent):
    if anaphor.attributes["type"] in ["PRO", "DEM", "VRB"]:
        return False
    elif antecedent.attributes["type"] in ["PRO", "DEM", "VRB"]:
        return False
    elif (anaphor.attributes["semantic_class"] == "NUMERIC" or
          antecedent.attributes["semantic_class"] == "NUMERIC"):
        return False
    else:
        return (anaphor.attributes["head"] != ["and"] and
                (" ".join(anaphor.attributes["head"]).lower()
                 == " ".join(antecedent.attributes["head"]).lower()))


def substring(anaphor, antecedent):
    if anaphor.attributes["type"] in ["PRO", "DEM", "VRB"]:
        return False
    elif antecedent.attributes["type"] != "NAM":
        return False
    elif (anaphor.attributes["semantic_class"] == "NUMERIC" or
          antecedent.attributes["semantic_class"] == "NUMERIC"):
        return False
    elif anaphor.attributes["head"] == ["and"]:
        return False
    else:
        cleaned = util.clean_via_pos(
            anaphor.attributes["tokens"],
            anaphor.attributes["pos"])

        return (" ".join(cleaned).lower()
                in " ".join(antecedent.attributes["tokens"]).lower())


def pronoun_same_canonical_form(anaphor, antecedent):
    return (anaphor.attributes["type"] == "PRO"
            and antecedent.attributes["type"] == "PRO"
            and (anaphor.attributes["citation_form"] ==
                 antecedent.attributes["citation_form"]))


def speaker(anaphor, antecedent):
    speaker_anaphor = anaphor.attributes["speaker"]
    speaker_antecedent = antecedent.attributes["speaker"]

    if speaker_anaphor == "-" and speaker_antecedent == "-":
        return False
    else:
        if (anaphor.attributes["type"] == "PRO"
                and antecedent.attributes["type"] == "PRO"):
            if (anaphor.attributes["citation_form"] == "i"
                    and antecedent.attributes["citation_form"] == "i"):
                return speaker_anaphor == speaker_antecedent
            elif ((anaphor.attributes["citation_form"] == "i"
                    and antecedent.attributes["citation_form"] == "you")
                  or (anaphor.attributes["citation_form"] == "you"
                      and antecedent.attributes["citation_form"] == "i")):
                return (nothing_between(anaphor, antecedent)
                        and speaker_anaphor != speaker_antecedent)
        elif (anaphor.attributes["type"] == "PRO"
              or antecedent.attributes["type"] == "PRO"):
            if (anaphor.attributes["type"] == "PRO"
                    and anaphor.attributes["citation_form"] == "i"):
                return (speaker_anaphor.replace("_", " ").lower() in
                        [" ".join(antecedent.attributes["tokens"]).lower(),
                         " ".join(antecedent.attributes["head"]).lower()])
            elif (antecedent.attributes["type"] == "PRO"
                    and antecedent.attributes["citation_form"] == "i"):
                return (speaker_antecedent.replace("_", " ").lower() in
                        [" ".join(anaphor.attributes["tokens"]).lower(),
                         " ".join(anaphor.attributes["head"]).lower()])


def nothing_between(anaphor, antecedent):
    if not anaphor.document:
        return True

    if anaphor.span < antecedent.span:
        start = anaphor.span.begin
        end = antecedent.span.end
    else:
        start = antecedent.span.begin
        end = anaphor.span.end

    speakers = anaphor.document.speakers[start:end+1]

    allowed_speakers = [speakers[0], speakers[-1]]
    for particular_speaker in speakers:
        if particular_speaker not in allowed_speakers:
            return False

    return True


def not_anaphoric(anaphor, antecedent):
    return not (anaphor.attributes["type"] in ["NAM", "PRO"]
                or (anaphor.attributes["type"] == "NOM"
                    and anaphor.attributes["fine_type"] == "DEF"))


def not_speaker(anaphor, antecedent):
    speaker_anaphor = anaphor.attributes["speaker"]
    speaker_antecedent = antecedent.attributes["speaker"]

    if speaker_anaphor == "-" or speaker_antecedent == "-":
        return False
    else:
        if (anaphor.attributes["type"] == "PRO"
                and antecedent.attributes["type"] == "PRO"):
            if ((anaphor.attributes["citation_form"] == "i"
                 and antecedent.attributes["citation_form"] == "i")
                or (anaphor.attributes["citation_form"] == "we"
                    and antecedent.attributes["citation_form"] == "we")
                or (anaphor.attributes["citation_form"] == "you"
                    and antecedent.attributes["citation_form"] == "you")):
                return speaker_anaphor != speaker_antecedent
            elif ((anaphor.attributes["citation_form"] == "i"
                   and antecedent.attributes["citation_form"] == "you")
                  or (anaphor.attributes["citation_form"] == "you"
                      and antecedent.attributes["citation_form"] == "i")):
                return speaker_anaphor == speaker_antecedent


def not_pronoun_distance(anaphor, antecedent):
    return (anaphor.attributes["type"] == "PRO"
            and anaphor.attributes["citation_form"] == "it"
            and (anaphor.attributes["sentence_id"]
                 - antecedent.attributes["sentence_id"] > 1))


def not_embedding(anaphor, antecedent):
    return (antecedent.span.embeds(anaphor.span)
            and (anaphor.attributes["fine_type"]
                 not in ["REFL", "POSS", "POSS_ADJ"]))


def not_compatible(anaphor, antecedent):
    if (" ".join(util.clean_via_pos(anaphor.attributes["tokens"],
                                    anaphor.attributes["pos"])).lower() ==
            " ".join(util.clean_via_pos(antecedent.attributes["tokens"],
                                        antecedent.attributes["pos"])).lower()):
        return False

    gender = (anaphor.attributes["gender"] == "UNKNOWN"
              or antecedent.attributes["gender"] == "UNKNOWN"
              or anaphor.attributes["gender"]
              == antecedent.attributes["gender"])

    number = (anaphor.attributes["number"] == "UNKNOWN"
              or antecedent.attributes["number"] == "UNKNOWN"
              or anaphor.attributes["number"]
              == antecedent.attributes["number"])

    semantic_class = (anaphor.attributes["semantic_class"] == "UNKNOWN"
                      or antecedent.attributes["semantic_class"] == "UNKNOWN"
                      or anaphor.attributes["semantic_class"]
                      == antecedent.attributes["semantic_class"])

    return not (gender and number and semantic_class)


def not_modifier(anaphor, antecedent):
    if (anaphor.attributes["type"] == "NAM"
            and antecedent.attributes["type"] == "NAM"):
        return False
    elif (anaphor.attributes["type"] in ["PRO", "DEM", "VRB"]
          or antecedent.attributes["type"] in ["PRO", "DEM", "VRB"]):
        return False
    else:
        return not get_modifier(anaphor).issubset(get_modifier(antecedent))


def get_modifier(mention):
    head_span_in_mention = spans.Span(
        mention.attributes["head_span"].begin - mention.span.begin,
        mention.attributes["head_span"].end - mention.span.begin)

    modifiers = set()

    for index, (token, pos) in enumerate(
            zip(mention.attributes["tokens"], mention.attributes["pos"])):
        if (token.lower() not in ["the", "this", "that", "those", "these",
                                  "a", "an"]
            and pos not in ["POS", "IN"]
            and (index < head_span_in_mention.begin
                 or index > head_span_in_mention.end)):
            modifiers.add(token.lower())

    return modifiers


def alias(anaphor, antecedent):
    if (anaphor.attributes["type"] != "NAM"
            or antecedent.attributes["type"] != "NAM"):
        return False
    elif (" ".join(anaphor.attributes["head"]).lower()
          == " ".join(antecedent.attributes["head"]).lower()):
        return False
    else:
        anaphor_cleaned_tokens = anaphor.attributes["head"]
        antecedent_cleaned_tokens = antecedent.attributes["head"]

        category = get_category_for_alias(
            anaphor.attributes["ner"][anaphor.attributes["head_index"]],
            antecedent.attributes["ner"][antecedent.attributes["head_index"]])

        if category == "PERSON":
            return person_alias(anaphor_cleaned_tokens,
                                antecedent_cleaned_tokens)
        elif category == "LOC":
            return loc_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens)
        elif category == "ORG":
            return org_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens)
        else:
            return False


def get_category_for_alias(anaphor_ner, antecedent_ner):
    if anaphor_ner == "PERSON" and antecedent_ner == "PERSON":
        return "PERSON"
    elif re.match(r"LOC", anaphor_ner) and re.match(r"LOC", antecedent_ner):
        return "LOC"
    elif re.match(r"ORG", anaphor_ner) and re.match(r"(ORG)", antecedent_ner):
        return "ORG"


def loc_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    return (starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens)
            or is_abbreviation(anaphor_cleaned_tokens,
                               antecedent_cleaned_tokens))


def org_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    return (starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens)
            or is_abbreviation(anaphor_cleaned_tokens,
                               antecedent_cleaned_tokens))


def person_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    if len(anaphor_cleaned_tokens) == 1 or len(antecedent_cleaned_tokens) == 1:
        return (anaphor_cleaned_tokens[0] == antecedent_cleaned_tokens[0]
                or anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1])
    elif (len(anaphor_cleaned_tokens) == 2
            and anaphor_cleaned_tokens[0].lower() in ["mr", "ms", "mr.", "ms."]
          or len(antecedent_cleaned_tokens) == 2
            and antecedent_cleaned_tokens[0].lower() in ["mr", "ms", "mr.",
                                                         "ms."]):
        return anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]
    elif (anaphor_cleaned_tokens[0] == antecedent_cleaned_tokens[0]
          and anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]):
        return True
    elif len(anaphor_cleaned_tokens) > 1 and len(antecedent_cleaned_tokens) > 1:
        return (anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]
                and anaphor_cleaned_tokens[-2] == antecedent_cleaned_tokens[-2])

    return False


def starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    for ana_token, ante_token in zip(anaphor_cleaned_tokens,
                                     antecedent_cleaned_tokens):
        if ana_token != ante_token:
            return False

    return True


def is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    if (" ".join(anaphor_cleaned_tokens).replace(".", "")
            == " ".join(antecedent_cleaned_tokens).replace(".", "")):
        return True
    else:
        if len(anaphor_cleaned_tokens) > len(antecedent_cleaned_tokens):
            return (" ".join(antecedent_cleaned_tokens)
                    in set(get_acronyms(anaphor_cleaned_tokens)))
        else:
            return (" ".join(anaphor_cleaned_tokens)
                    in set(get_acronyms(antecedent_cleaned_tokens)))


def get_acronyms(cleaned_tokens):
    company_designator = r'assoc|bros|co|coop|corp|devel|inc|llc|ltd\.?'
    tokens_without_designator = [token for token in cleaned_tokens
                                 if not re.match(company_designator,
                                                 token.lower())]

    return (" ".join(tokens_without_designator),
            "".join([token[0] for token in tokens_without_designator
                    if token[0].isupper()]),
            ".".join([token[0] for token in tokens_without_designator
                      if token[0].isupper()])+".")



""" Extract coreference information from pairwise predictions."""

__author__ = 'smartschat'


def best_first(substructures, labels, scores, coref_labels):
    """ Extract coreference clusters from coreference predictions via best-first
    clustering.

    In particular, go through a list of anaphor-antecedent pairs, where
    pairs with the same anaphor are consecutive. Then, for each anaphor, the
    best-scoring antecedent is selected (this is also called best-first
    clustering). Ties are broken by position in the list: earlier items are
    preferred.

    Args:
        substructures (list(list((Mention, Mention)))): A list of substructures.
            For this clusterer, each substructure should contain only one
            (anaphor, antecedent) pair. If two substructures have the same
            anaphor, they should be consecutive.
        labels (list(list(str))): A list of arc labels. This list should
            have the same length as the list of substructures, and each inner
            list should contain only one element (as in ``substructures``).
            Each entry describes the label of an arc.
        labels (list(list(str))): A list of arc scores. This list should
            have the same length as the list of substructures, and each inner
            list should contain only one element (as in ``substructures``).
            Each entry describes the score of an arc.
        coref_labels (set(str)): A list of labels that indicate that mentions
            connected via an arc that has one of these labels are coreferent.

    Returns
        A tuple containing two dicts. The components are

            - **mention_entity_mapping** (*dict(Mention, int)*): A mapping of
              mentions to entity identifiers.
            - **antecedent_mapping** (*dict(Mention, Mention)*): A mapping of
              mentions to their antecedent.
    """

    anaphor = None
    best = None
    max_val = float('-inf')

    mention_entity_mapping = {}
    antecedent_mapping = {}

    for substructure, substructure_label, substructure_score in zip(
            substructures, labels, scores):
        # each substructure consists of one pair
        pair = substructure[0]
        label = substructure_label[0]
        score = substructure_score[0]
        current_anaphor, current_antecedent = pair
        if current_anaphor != anaphor:
            # change in anaphor: set coreference information based on
            # best-scoring antecedent
            if anaphor and best and not best.is_dummy():
                antecedent_mapping[anaphor] = best
                if best not in mention_entity_mapping:
                    mention_entity_mapping[best] = \
                        best.document.system_mentions.index(best)

                mention_entity_mapping[anaphor] = \
                    mention_entity_mapping[best]

            best = None
            max_val = float('-inf')

        if score > max_val and label in coref_labels:
            max_val = score
            best = current_antecedent

        anaphor = current_anaphor

    if anaphor and best and not best.is_dummy():
        antecedent_mapping[anaphor] = best
        if best not in mention_entity_mapping:
            mention_entity_mapping[best] = \
                best.document.system_mentions.index(best)

        mention_entity_mapping[anaphor] = \
            mention_entity_mapping[best]

    return mention_entity_mapping, antecedent_mapping


def all_ante(substructures, labels, scores, coref_labels):
    """ Extract coreference clusters from coreference predictions via transitive
    closure.

    In particular, go through all (anaphor, antecedent) pairs contained in
    ``substructures``, and obtain coreference clusters by transitive closure.

    Args:
        substructures (list(list((Mention, Mention)))): A list of substructures.
        labels (list(list(str))): Not used by this function.
        labels (list(list(str))): Not used by this function.
        coref_labels (set(str)): Not used by this function.

    Returns
        A tuple containing two dicts. The components are

            - **mention_entity_mapping** (*dict(Mention, int)*): A mapping of
              mentions to entity identifiers.
            - **antecedent_mapping** (*dict(Mention, Mention)*): A mapping of
              mentions to their antecedent.
    """
    mention_entity_mapping = {}
    antecedent_mapping = {}

    for substructure in substructures:
        for pair in substructure:
            anaphor, antecedent = pair

            # skip dummy antecedents
            if antecedent.is_dummy():
                continue

            antecedent_mapping[anaphor] = antecedent

            # antecedent is not in the mapping: we initialize a new coreference
            # chain
            if antecedent not in mention_entity_mapping:
                # chain id: index of antecedent in system mentions
                mention_entity_mapping[antecedent] = \
                        antecedent.document.system_mentions.index(antecedent)

            # assign id based on antecedent
            mention_entity_mapping[anaphor] = \
                mention_entity_mapping[antecedent]

    return mention_entity_mapping, antecedent_mapping



import unittest


import nltk


from cort.core.spans import Span
from cort.core.documents import CoNLLDocument
from cort.core import mention_property_computer


__author__ = 'smartschat'


class TestMentionPropertyComputer(unittest.TestCase):
    def test_number(self):
        self.assertEqual(
            "SINGULAR",
            mention_property_computer.compute_number(
                {"tokens": ["him"], "pos": ["PRP"], "type": "PRO",
                 "citation_form": "he", "head_index": 0}))
        self.assertEqual(
            "SINGULAR",
            mention_property_computer.compute_number(
                {"tokens": ["the", "guy"], "pos": ["DT", "NN"], "type": "NOM",
                 "head_index": 1}))
        self.assertEqual(
            "PLURAL",
            mention_property_computer.compute_number(
                {"tokens": ["they"], "pos": ["PRP"],"type": "PRO",
                 "citation_form": "they", "head_index": 0}))
        self.assertEqual(
            "PLURAL",
            mention_property_computer.compute_number(
                {"tokens": ["these", "freaks"], "pos": ["DT", "NNS"],
                 "type": "NOM", "head_index": 1}))
        self.assertEqual(
            "PLURAL",
            mention_property_computer.compute_number(
                {"tokens": ["he", "and", "she"], "pos": ["PRP", "CC", "PRP"],
                 "type": "NOM", "head_index": 1}))

    def test_gender(self):
        self.assertEqual(
            "MALE",
            mention_property_computer.compute_gender(
                {"tokens": ["him"], "pos": ["PRP"], "type": "PRO",
                 "citation_form": "he", "head_index": 0}))
        self.assertEqual(
            "NEUTRAL",
            mention_property_computer.compute_gender(
                {"tokens": ["its"], "pos": ["PRP$"], "type": "PRO",
                 "citation_form": "it", "head_index": 0}))
        self.assertEqual(
            "FEMALE",
            mention_property_computer.compute_gender(
                {"tokens": ["Mrs.", "Robinson"], "pos": ["NNP", "NNP"],
                 "type": "NAM", "head_index": 1}))
        self.assertEqual(
            "MALE",
            mention_property_computer.compute_gender(
                {"tokens": ["Mr.", "FooBar"], "pos": ["NNP", "NNP"],
                 "type": "NAM", "head_index": 1}))
        self.assertEqual(
            "NEUTRAL",
            mention_property_computer.compute_gender(
                {"tokens": ["an", "arrow"], "head": ["arrow"],
                 "pos": ["DT", "NN"],"type": "NOM", "ner": ["-", "-"],
                 "head_index": 1}))
        self.assertEqual(
            "FEMALE",
            mention_property_computer.compute_gender(
                {"tokens": ["the", "girl"], "head": ["girl"],
                 "pos": ["DT", "NN"],"type": "NOM", "ner": ["-", "-"],
                 "head_index": 1}))
        self.assertEqual(
            "NEUTRAL",
            mention_property_computer.compute_gender(
                {"tokens": ["the", "shooting"], "head": ["shooting"],
                 "pos": ["DT", "NN"],"type": "NOM", "ner": ["-", "-"],
                 "head_index": 1}))
        self.assertEqual(
            "MALE",
            mention_property_computer.compute_gender(
                {"tokens": ["the", "groom"], "head": ["groom"],
                 "pos": ["DT", "NN"],"type": "NOM", "ner": ["-", "-"],
                 "head_index": 1}))
        self.assertEqual(
            "PLURAL",
            mention_property_computer.compute_gender(
                {"tokens": ["the", "guys"], "head": ["guys"],
                 "pos": ["DT", "NNS"],"type": "NOM", "ner": ["-", "-"],
                 "head_index": 1}))
        self.assertEqual(
            "NEUTRAL",
            mention_property_computer.compute_gender(
                {"tokens": ["the", "Mona", "Lisa"], "head": ["Mona", "Lisa"],
                 "pos": ["DT", "NNP", "NNP"],"type": "NAM",
                 "ner": ["-", "WORK_OF_ART", "WORK_OF_ART"], "head_index": 2}))

    def test_semantic_class(self):
        self.assertEqual(
            "PERSON",
            mention_property_computer.compute_semantic_class(
                {"tokens": ["him"], "pos": ["PRP"], "type": "PRO",
                 "citation_form": "he", "head_index": 0}))
        self.assertEqual(
            "OBJECT",
            mention_property_computer.compute_semantic_class(
                {"tokens": ["its"], "pos": ["PRP$"], "type": "PRO",
                 "citation_form": "it", "head_index": 0}))
        self.assertEqual(
            "PERSON",
            mention_property_computer.compute_semantic_class(
                {"tokens": ["Mrs.", "Robinson"], "ner": ["PERSON", "PERSON"],
                 "pos": ["NNP", "NNP"],"type": "NAM", "head_index": 1}))
        self.assertEqual(
            "OBJECT",
            mention_property_computer.compute_semantic_class(
                {"tokens": ["a", "house"], "head": ["house"],
                 "pos": ["DT", "NN"],"type": "NOM", "ner": ["NONE", "NONE"],
                 "head_index": 1}))
        self.assertEqual(
            "UNKNOWN",
            mention_property_computer.compute_semantic_class(
                {"tokens": ["adsfg"], "head": ["adsfg"],
                 "pos": ["NN"],"type": "NOM", "ner": ["NONE", "NONE"],
                 "head_index": 0}))

    def test_citation_form(self):
        self.assertEqual(
            "they",
            mention_property_computer.get_citation_form({"tokens": ["them"]}))
        self.assertEqual(
            "she",
            mention_property_computer.get_citation_form({"tokens": ["her"]}))
        self.assertEqual(
            None,
            mention_property_computer.get_citation_form({"tokens": ["why"]}))

    def test_get_head_index(self):
        self.real_example = """#begin document (bn/voa/02/voa_0220); part 000
bn/voa/02/voa_0220   0    0    Unidentified    JJ  (TOP(S(NP(NP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    1          gunmen   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    2              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    3           north    JJ      (NP(ADJP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    4         western    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    5        Colombia   NNP            *)))         -   -   -   -         (GPE)   -
bn/voa/02/voa_0220   0    6            have   VBP           (VP*        have  -   -   -            *    -
bn/voa/02/voa_0220   0    7       massacred   VBN           (VP*    massacre  -   -   -            *    -
bn/voa/02/voa_0220   0    8              at    IN   (NP(QP(ADVP*          -   -   -   -   (CARDINAL*    -
bn/voa/02/voa_0220   0    9           least   JJS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   10          twelve    CD              *)         -   -   -   -            *)   -
bn/voa/02/voa_0220   0   11        peasants   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   12              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   13             the    DT        (NP(NP*          -   -   -   -            *   (0
bn/voa/02/voa_0220   0   14          second    JJ              *          -   -   -   -     (ORDINAL)   -
bn/voa/02/voa_0220   0   15            such    JJ              *          -   -   -   -            *    -
bn/voa/02/voa_0220   0   16        incident    NN              *)   incident  -   2   -            *    -
bn/voa/02/voa_0220   0   17              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   18              as    RB        (NP(QP*          -   -   -   -       (DATE*    -
bn/voa/02/voa_0220   0   19            many    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   20            days   NNS         *))))))        day  -   4   -            *)   0)
bn/voa/02/voa_0220   0   21               .     .             *))         -   -   -   -            *    -

bn/voa/02/voa_0220   0    0          Local    JJ    (TOP(S(NP*          -    -   -   -   *   (ARG0*             *    -
bn/voa/02/voa_0220   0    1         police   NNS             *)     police   -   -   -   *        *)            *    -
bn/voa/02/voa_0220   0    2            say   VBP          (VP*         say  01   1   -   *      (V*)            *    -
bn/voa/02/voa_0220   0    3             it   PRP   (SBAR(S(NP*)         -    -   -   -   *   (ARG1*        (ARG1*)   -
bn/voa/02/voa_0220   0    4             's   VBZ          (VP*          be  01   1   -   *        *           (V*)   -
bn/voa/02/voa_0220   0    5            not    RB             *          -    -   -   -   *        *    (ARGM-NEG*)   -
bn/voa/02/voa_0220   0    6          clear    JJ        (ADJP*)         -    -   -   -   *        *        (ARG2*)   -
bn/voa/02/voa_0220   0    7            who    WP   (SBAR(WHNP*)         -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0    8            was   VBD        (S(VP*          be   -   1   -   *        *             *    -
bn/voa/02/voa_0220   0    9    responsible    JJ        (ADJP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   10            for    IN          (PP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   11            the    DT          (NP*          -    -   -   -   *        *             *   (0
bn/voa/02/voa_0220   0   12       massacre    NN    *))))))))))   massacre   -   -   -   *        *)            *    0)
bn/voa/02/voa_0220   0   13              .     .            *))         -    -   -   -   *        *             *    -

#end document
"""

        real_document = CoNLLDocument(self.real_example)

        expected = 0
        head = nltk.ParentedTree.fromstring("(WHNP (WP who))")
        mention_subtree = mention_property_computer.get_relevant_subtree(
            Span(29, 34), real_document)
        self.assertEqual(expected, mention_property_computer.get_head_index(
            head, mention_subtree))

    def test_tree_is_apposition(self):
        self.assertEqual(
            True,
            mention_property_computer.is_apposition({
                "parse_tree": nltk.ParentedTree.fromstring(
                    "(NP (NP (NP (NNP Secretary)) (PP (IN of) (NP "
                    "(NNP State)))) (NP (NNP Madeleine) (NNP Albright)))")}))
        self.assertEqual(
            False,
            mention_property_computer.is_apposition({
                "parse_tree": nltk.ParentedTree.fromstring(
                    "(NP (NP (NNP Secretary)) (PP (IN of) "
                    "(NP (NNP State))))")}))
        self.assertEqual(
            False,
            mention_property_computer.is_apposition({
                "parse_tree": nltk.ParentedTree.fromstring(
                    "(NP (NP (NP (NNP Al) (NNP Gore) (POS 's)) (NN campaign) "
                    "(NN manager)) (, ,) (NP (NNP Bill) (NNP Daley)) (, ,))")}))
        self.assertEqual(
            False,
            mention_property_computer.is_apposition({
                "parse_tree": nltk.ParentedTree.fromstring(
                    "(NP (NP (NNS news)) (NP (CD today)))")}))
        self.assertEqual(
            False,
            mention_property_computer.is_apposition({
                "parse_tree": nltk.ParentedTree.fromstring(
                    "(NP (NP (PRP$ his) (NN brother)) (, ,) (NP (PRP$ his) "
                    "(NN sister)))")}))
        self.assertEqual(
            True,
            mention_property_computer.is_apposition({
                "parse_tree":  nltk.ParentedTree.fromstring(
                    "(NP (NP (NNP Barack) (NNP Obama)) (, ,) (NP (DT the) "
                    "(NN president)))")}))

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import codecs
import shutil
import html
import os
import webbrowser
from random import randint
import collections

import cort
from cort.core import spans
from cort.analysis import data_structures


__author__ = "Thierry Goeckel, Sebastian Martschat"


class Visualizer:
    def __init__(self, structured_coreference_analysis, corpus_name,
                 highlight_error=None, for_raw_input=False):
        self.html_header = \
            "<!doctype html>\n" \
            "<html>\n" \
            "\t<head>\n" \
            "\t\t<title>cort visualization</title>\n" \
            "\t\t<link rel=\"stylesheet\" type=\"text/css\" " \
            "href=\"visualization/style.css\">\n" \
            "\t\t<script src=\"visualization/lib/jquery-2.1.1.min.js\">" \
            "</script>\n" \
            "\t\t<script src=\"visualization/lib/jquery.jsPlumb-1.6.4.js\"></script>\n" \
            "\t\t<script src=\"visualization/lib/cort.js\"></script>"

        self.chain_to_colour = {}
        self.colours = []
        self.navi = {}
        self.structured_coreference_analysis = structured_coreference_analysis
        self.highlight_error = highlight_error
        self.corpus_name = corpus_name
        self.for_raw_input = for_raw_input

    def run(self):
        documents_html = ""
        documents_navi = "\n\t\t<div id=\"documentsNavi\"><h3>Documents</h3>" \
                         "\n\t\t\t<ul>"
        errors_source = "\n\t\t<script>\n" \
                        "\t\t\terrors = [ "

        system_corpus = self.structured_coreference_analysis.corpora[
            self.corpus_name]

        for document in sorted(self.structured_coreference_analysis.reference.documents,
                               key=lambda doc:
                               doc.get_html_friendly_identifier()):
            doc_id = document.get_html_friendly_identifier()
            documents_navi += "\n\t\t\t\t<li>" + doc_id + "</li>"

            document_mentions = document.annotated_mentions

            if not self.for_raw_input:
                system_mentions = system_corpus.documents[
                                        system_corpus.documents.index(document)
                                    ].annotated_mentions
                document_mentions = sorted(document_mentions + system_mentions)

            if self.for_raw_input:
                text_source = self.__generate_html_for_raw(document,
                                                           document_mentions)
            else:
                text_source = self.__generate_html_for_errors(document,
                                                              document_mentions)

            if self.for_raw_input:
                documents_html = documents_html + "\n\t\t<div id=\"" + doc_id + \
                                "\" class=\"document\">" + \
                                "\n\t\t\t<div class=\"navcontainer\">" + \
                                self.navi["system"] + \
                                "\n\t\t\t\t\t</ul>\n\t\t\t\t</div>" \
                                "\n\t\t\t</div>" + \
                                text_source + "\n\t\t</div>"

            else:
                documents_html = documents_html + "\n\t\t<div id=\"" + doc_id + \
                                "\" class=\"document\">" \
                                "\n\t\t\t<div class=\"navcontainer\">" + \
                                self.__generate_errors_navi_by_mention_type(document) + \
                                self.navi["gold"] + \
                                "\n\t\t\t\t\t</ul>\n\t\t\t\t</div>" + \
                                self.navi["system"] + \
                                "\n\t\t\t\t\t</ul>\n\t\t\t\t</div>" \
                                "\n\t\t\t</div>" + \
                                text_source + \
                                "\n\t\t</div>"

            recall_errors = self.structured_coreference_analysis[
                self.corpus_name]["recall_errors"]["all"].filter(
                            lambda err: err[0].document == document)
            precision_errors = self.structured_coreference_analysis[
                self.corpus_name]["precision_errors"]["all"].filter(
                            lambda err: err[0].document == document)
            decisions = self.structured_coreference_analysis[
                self.corpus_name]["decisions"]["all"].filter(
                            lambda err: err[0].document == document)

            if isinstance(recall_errors, data_structures.StructuredCoreferenceAnalysis):
                for category in recall_errors.keys():
                    errors_source += self.__generate_errors_source(
                        recall_errors[category],
                        category,
                        document_mentions,
                        "Recall")
            else:
                errors_source += self.__generate_errors_source(
                    recall_errors,
                    "",
                    document_mentions,
                    "Recall")

            if isinstance(precision_errors, data_structures.StructuredCoreferenceAnalysis):
                for category in precision_errors.keys():
                    errors_source += self.__generate_errors_source(
                        precision_errors[category],
                        category,
                        document_mentions,
                        "Precision")
            else:
                errors_source += self.__generate_errors_source(
                    precision_errors,
                    "",
                    document_mentions,
                    "Precision")

            if isinstance(decisions, data_structures.StructuredCoreferenceAnalysis):
                for category in decisions.keys():
                    errors_source += self.__generate_errors_source(
                        decisions[category],
                        category,
                        document_mentions,
                        "Decision")
            else:
                errors_source += self.__generate_errors_source(
                    decisions,
                    "",
                    document_mentions,
                    "Decision")

        errors_source = errors_source[:-2] + " ];\n\t\t\tchain_to_colour = {"

        for chain in self.chain_to_colour.keys():
            errors_source = errors_source + chain + ": \"" + self.chain_to_colour[chain] + "\", "

        errors_source = errors_source[:-2] + "};\n\t\t</script>"

        html_source = self.html_header + \
            "\n\t</head>\n\t<body>" \
            "\n\t\t<div id=\"header\"><h1>cort visualization: <span id=\"document_name\">" \
            "</span></h1></div>" + documents_navi + \
            "\n\t\t\t</ul>\n\t\t</div>"

        html_source += documents_html + errors_source + "\n\t</body>\n</html>"

        if not os.path.exists("temp/output"):
            os.makedirs("temp/output")

        output = "temp/output/error_analysis.html"

        f = codecs.open(output, "w", "utf-8")

        abs_path = os.path.abspath(output)

        print("Writing " + abs_path)

        f.write(html_source)

        f.close()

        # copy js/css
        package_dir = cort.__path__[0]
        if not os.path.exists("temp/output/visualization"):
            shutil.copytree(package_dir + "/analysis/visualization",
                            "temp/output/visualization")

        if self.for_raw_input:
            shutil.copy(package_dir + "/analysis/visualization/lib/cort-for-raw.js",
                        "temp/output/visualization/lib/cort.js")

        webbrowser.open_new_tab("file://" + abs_path)

    def __generate_html_for_errors(self, document, mentions):
        document_html = "\n\t\t\t<ol class=\"text\">\n" \
                        "\t\t\t\t<li class=\"sentence\">"

        self.navi["gold"] = "\n\t\t\t\t<div class=\"goldNavi\">" \
                            "<h3>Reference Entities</h3>" \
                            "<span class=\"tease\">show all</span>" \
                            "\n\t\t\t\t\t<ul>"

        self.navi["system"] = "\n\t\t\t\t<div class=\"systemNavi\">" \
                              "<h3>System Entities</h3>" \
                              "<span class=\"tease\">show all</span>" \
                              "\n\t\t\t\t\t<ul>"

        chains = set()

        index = 0

        sentence_id, sentence_span = document.get_sentence_id_and_span(
            spans.Span(0, 0))

        annotated_mentions = set(document.annotated_mentions)

        for token in document.tokens:
            token = html.escape(token)

            mention_id = 0

            mention_text = ""

            processed_gold_mentions = set()

            for mention in mentions:
                if mention.span.begin > index:
                    break

                if mention.span.end < index:
                    mention_id += 1
                    continue

                mention_tokens = html.escape(" ".join(mention.attributes[
                    'tokens']), True)

                mention_head = html.escape(" ".join(mention.attributes[
                    'head']), True)

                mention_type = html.escape("".join(mention.attributes[
                    'type']), True)

                mention_span = str(mention.span)

                if mention in annotated_mentions and \
                   mention not in processed_gold_mentions:
                    system = "gold"

                    processed_gold_mentions.add(mention)
                else:
                    system = "system"

                chain_id = system + str(mention.attributes['annotated_set_id'])

                if chain_id not in chains:
                    self.navi[system] += "\n\t\t\t\t\t\t<li class=\"" + \
                                         chain_id +\
                                         "\">" + mention_tokens + "</li>"
                    chains.add(chain_id)

                if chain_id not in self.chain_to_colour.keys():
                    while True:
                        r = lambda: randint(170, 255)
                        colour = '#%02X%02X%02X' % (r(), r(), r())
                        if colour not in self.colours:
                            self.colours.append(colour)
                            break

                    self.chain_to_colour[chain_id] = colour

                span_id = document.get_html_friendly_identifier() + "_" + \
                          str(mention_id)

                temp_text = "<span " \
                            "id=\"" + span_id + "\" " \
                            "class=\"" + chain_id + " mention\" " \
                            "data-mentiontype=\"" + mention_type + "\" " \
                            "data-mentionhead=\"" + mention_head + "\" " \
                            "data-span=\"" + mention_span + "\">"

                if mention.span.begin == index and mention.span.end == index:
                    if mention_text.endswith("</span> "):
                        mention_text = temp_text + mention_text.strip() + \
                            "</span> "
                    elif mention_text == "":
                        mention_text = temp_text + token + "</span> "
                elif mention.span.begin == index:
                    if mention_text == "":
                        mention_text = temp_text + token + " "
                    else:
                        mention_text = temp_text + mention_text
                elif mention.span.end == index:
                    if mention_text == "":
                        mention_text = token + "</span> "
                    else:
                        mention_text = mention_text.strip() + "</span> "

                mention_id += 1

            if mention_text == "":
                mention_text = token + " "

            token_span = spans.Span(index, index)

            if document.get_sentence_id_and_span(token_span) is None or \
                    sentence_span != document.get_sentence_id_and_span(
                            token_span)[1]:
                mention_text = "</li>\n" \
                               "\t\t\t\t<li class=\"sentence\">" + mention_text

                sentence_id, sentence_span = document.get_sentence_id_and_span(token_span)

            document_html += mention_text

            index += 1

        document_html.strip()

        return document_html + "</li>\n\t\t\t</ol>"

    def __generate_html_for_raw(self, document, mentions):
        document_html = "\n\t\t\t<ol class=\"text\">\n" \
                        "\t\t\t\t<li class=\"sentence\">"

        self.navi["gold"] = "\n\t\t\t\t<div class=\"goldNavi\">" \
                            "<h3>Reference Entities</h3>" \
                            "<span class=\"tease\">show all</span>" \
                            "\n\t\t\t\t\t<ul>"

        self.navi["system"] = "\n\t\t\t\t<div class=\"systemNavi\">" \
                              "<h3>System Entities</h3>" \
                              "<span class=\"tease\">show all</span>" \
                              "\n\t\t\t\t\t<ul>"

        chains = set()

        index = 0

        sentence_id, sentence_span = document.get_sentence_id_and_span(
            spans.Span(0, 0))

        annotated_mentions = set(document.annotated_mentions)

        temp_navi = {
            "gold": {},
            "system": {}
        }

        chain_counter = {
            "gold": collections.Counter(),
            "system": collections.Counter()
        }

        for system in ["gold", "system"]:
            for mention in annotated_mentions:
                chain_counter[system].update([system + str(mention.attributes[
                    "annotated_set_id"])])

        for token in document.tokens:
            token = html.escape(token)

            mention_id = 0

            mention_text = ""

            processed_gold_mentions = set()

            for mention in mentions:
                if mention.span.begin > index:
                    break

                if mention.span.end < index:
                    mention_id += 1
                    continue

                mention_tokens = html.escape(" ".join(mention.attributes[
                    'tokens']), True)

                mention_head = html.escape(" ".join(mention.attributes[
                    'head']), True)

                mention_type = html.escape("".join(mention.attributes[
                    'type']), True)

                mention_span = str(mention.span)

                system = "system"

                chain_id = system + str(mention.attributes['annotated_set_id'])

                if chain_id not in chains:
                    temp_navi[system][chain_id] = "\n\t\t\t\t\t\t<li " \
                                                  "class=\"" + \
                                         chain_id +\
                                         "\">" + mention_tokens + "</li>"
                    chains.add(chain_id)

                if chain_id not in self.chain_to_colour.keys():
                    while True:
                        r = lambda: randint(170, 255)
                        colour = '#%02X%02X%02X' % (r(), r(), r())
                        if colour not in self.colours:
                            self.colours.append(colour)
                            break

                    self.chain_to_colour[chain_id] = colour

                span_id = document.get_html_friendly_identifier() + "_" + \
                          str(mention_id)

                style = ""

                if chain_counter[system][chain_id] > 1:
                    style = "style=\"background-color:" + self.chain_to_colour[
                        chain_id] + "\" "

                temp_text = "<span " \
                            "id=\"" + span_id + "\" " \
                            "class=\"" + chain_id + " mention\" "  + \
                            style + \
                            "data-mentiontype=\"" + mention_type + "\" " \
                            "data-mentionhead=\"" + mention_head + "\" " \
                            "data-span=\"" + mention_span + "\">"

                if mention.span.begin == index and mention.span.end == index:
                    if mention_text.endswith("</span> "):
                        mention_text = temp_text + mention_text.strip() + \
                            "</span> "
                    elif mention_text == "":
                        mention_text = temp_text + token + "</span> "
                elif mention.span.begin == index:
                    if mention_text == "":
                        mention_text = temp_text + token + " "
                    else:
                        mention_text = temp_text + mention_text
                elif mention.span.end == index:
                    if mention_text == "":
                        mention_text = token + "</span> "
                    else:
                        mention_text = mention_text.strip() + "</span> "

                mention_id += 1

            if mention_text == "":
                mention_text = token + " "

            token_span = spans.Span(index, index)

            if document.get_sentence_id_and_span(token_span) is None or \
                    sentence_span != document.get_sentence_id_and_span(
                            token_span)[1]:
                mention_text = "</li>\n" \
                               "\t\t\t\t<li class=\"sentence\">" + mention_text

                sentence_id, sentence_span = document.get_sentence_id_and_span(token_span)

            document_html += mention_text

            index += 1

        document_html.strip()

        for system in ["system"]:
            for key, val in chain_counter[system].items():
                if val > 1:
                    self.navi[system] += temp_navi[system][key]

        return document_html + "</li>\n\t\t\t</ol>"

    def __generate_errors_source(self, errors, category, sorted_mentions,
                                 error_type):
        errors_source = ""

        for error in errors:
            doc_id = error[0].document.get_html_friendly_identifier()
            antecedent_id = -1
            anaphor_id = -1
            mention_id = 0
            for mention in sorted_mentions:
                if mention == error[1] and antecedent_id == -1:
                    antecedent_id = mention_id
                if mention == error[0] and anaphor_id == -1:
                    anaphor_id = mention_id
                if antecedent_id > -1 and anaphor_id > -1:
                    break
                mention_id += 1
            error_source = \
                "{ anaphor: \"" + doc_id + "_" + str(anaphor_id) + "\", " \
                "antecedent: \"" + doc_id + "_" \
                + str(antecedent_id) + "\", " \
                "category: \"" + str(category) + "\", " \
                "type: \"" + error_type + "\""
            if error == self.highlight_error:
                error_source += ", highlight: \"true\" }, "
            else:
                error_source += " }, "
            errors_source += error_source

        return errors_source

    def __generate_errors_navi_by_mention_type(self, document):
        recall_errors = self.structured_coreference_analysis[
                            self.corpus_name]["recall_errors"]["all"].filter(
            lambda x: x[0].document == document)
        precision_errors = self.structured_coreference_analysis[
                            self.corpus_name]["precision_errors"]["all"].filter(
            lambda x: x[0].document == document)

        recall_error_count = len(recall_errors)

        precision_error_count = len(precision_errors)

        # Precision errors
        precision_errors_navi = "\n\t\t\t\t\t<div><h4>Precision (" + \
                            str(precision_error_count) +\
                            ")</h4><span class=\"tease\">show all</span>" \
                            "\n\t\t\t\t\t\t<ul class=\"precisionErrors\">"

        # Recall errors
        recall_errors_navi = "\n\t\t\t\t\t\t<div><h4>Recall (" + \
                            str(recall_error_count) + \
                            ")</h4><span class=\"tease\">show all</span>" \
                            "\n\t\t\t\t\t\t<ul class=\"recallErrors\">"

        recall_categories = set()
        precision_categories = set()

        if isinstance(recall_errors, data_structures.StructuredCoreferenceAnalysis):
            for cat in recall_errors.keys():
                if len(recall_errors[cat]) > 0:
                    recall_categories.add(cat)

        if isinstance(precision_errors, data_structures.StructuredCoreferenceAnalysis):
            for cat in precision_errors.keys():
                if len(precision_errors[cat]) > 0:
                    precision_categories.add(cat)

        recall_categories = sorted([str(cat) for cat in recall_categories])
        precision_categories = sorted([str(cat) for cat
                                       in precision_categories])

        for cat in recall_categories:
            recall_errors_navi += "\n\t\t\t\t\t\t\t<li>"

            error_count_cat = len(recall_errors[cat])

            recall_errors_navi += str(cat) + ": " + \
                str(error_count_cat)

            recall_errors_navi += "</li>"

        for cat in precision_categories:
            precision_errors_navi += "\n\t\t\t\t\t\t\t<li>"

            error_count_cat = len(precision_errors[cat])

            precision_errors_navi += str(cat) + ": " + \
                str(error_count_cat)

            precision_errors_navi += "</li>"

        precision_errors_navi += "\n\t\t\t\t\t\t</ul>" \
                            "\n\t\t\t\t\t</div>"
        recall_errors_navi += "\n\t\t\t\t\t\t</ul>" \
                            "\n\t\t\t\t\t</div>"
        return "\n\t\t\t\t<div class=\"errorsNavi\">" \
                "\n\t\t\t\t\t<h3>Errors (" + \
                str(precision_error_count + recall_error_count) + ")</h3>" \
                "\n\t\t\t\t\t<span class=\"tease\">show all</span>" + \
                precision_errors_navi + recall_errors_navi + \
                "\n\t\t\t\t</div>"



from collections import namedtuple
import unittest

from cort.analysis import data_structures
from cort.analysis import error_extractors
from cort.analysis import spanning_tree_algorithms
from cort.core import corpora
from cort.core import mentions
from cort.core import spans

__author__ = 'smartschat'


class TestErrorExtractor(unittest.TestCase):
    def setUp(self):
        self.first_cluster = [
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": ["a"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(1, 1),
                {"tokens": ["b"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(2, 3),
                {"tokens": ["c", "d"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(4, 5),
                {"tokens": ["e", "f"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": ["f", "g"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(7, 7),
                {"tokens": ["h"], "annotated_set_id": 0}),
        ]

        self.second_cluster = [
            mentions.Mention(
                None,
                spans.Span(3, 4),
                {"tokens": ["d", "e"], "annotated_set_id": 1}),

            mentions.Mention(
                None,
                spans.Span(7, 8),
                {"tokens": ["h", "i"], "annotated_set_id": 1}),

            mentions.Mention(
                None,
                spans.Span(10, 10),
                {"tokens": ["k"], "annotated_set_id": 1})
        ]

        self.system_cluster = [
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": ["a"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(2, 3),
                {"tokens": ["c", "d"], "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(4, 5),
                {"tokens": ["e", "f"], "annotated_set_id": 2}),

            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": ["f", "g"], "annotated_set_id": 2}),

            mentions.Mention(
                None,
                spans.Span(7, 7),
                {"tokens": ["h"], "annotated_set_id": 1}),

            mentions.Mention(
                None,
                spans.Span(10, 10),
                {"tokens": ["k"], "annotated_set_id": 1})
        ]

        self.maxDiff = None

    def test_compute_errors(self):
        # fake document using a named tuple
        document = namedtuple("Document", "annotated_mentions")
        doc_gold = document(self.first_cluster + self.second_cluster)
        doc_system = document(self.system_cluster)
        corpus_gold = corpora.Corpus("fake gold", [doc_gold])
        corpus_system = corpora.Corpus("fake system", [doc_system])

        ex = error_extractors.ErrorExtractor(
            corpus_gold,
            spanning_tree_algorithms.recall_closest,
            spanning_tree_algorithms.precision_system_output
        )

        ex.add_system(corpus_system)

        self.assertEqual(
            data_structures.EnhancedSet([
                (self.first_cluster[1], self.first_cluster[0]),
                (self.first_cluster[3], self.first_cluster[2]),
                (self.first_cluster[5], self.first_cluster[4]),
                (self.second_cluster[1], self.second_cluster[0]),
                (self.second_cluster[2], self.second_cluster[1]),
            ]),
            ex.get_errors()["fake system"]["recall_errors"]["all"]
        )

    if __name__ == '__main__':
        unittest.main()



""" Compute attributes of mentions. """

import re

from nltk.corpus import wordnet as wn

from cort.core import external_data
from cort.core import head_finders
from cort.core import spans


__author__ = 'smartschat'


def compute_number(attributes):
    """ Compute the number of a mention.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            values for "type", "head_index" and "pos".

    Returns:
        str: the number of the mention -- one of UNKNOWN, SINGULAR and PLURAL.
    """
    number = "UNKNOWN"
    head_index = attributes["head_index"]
    pos = attributes["pos"][head_index]

    if attributes["type"] == "PRO":
        if attributes["citation_form"] in ["i", "you", "he", "she", "it"]:
            number = "SINGULAR"
        else:
            number = "PLURAL"
    elif attributes["type"] == "DEM":
        if attributes["head"][0].lower() in ["this", "that"]:
            number = "SINGULAR"
        else:
            number = "PLURAL"
    elif attributes["type"] in ["NOM", "NAM"]:
        if pos == "NNS" or pos == "NNPS":
            number = "PLURAL"
        else:
            number = "SINGULAR"

    if pos == "CC":
        number = "PLURAL"

    return number


def compute_gender(attributes):
    """ Compute the gender of a mention.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            values for "type", "head", "head_index" and, if the mention is a
            pronoun, "citation_form".

    Returns:
        str: the number of the mention -- one of UNKNOWN, MALE, FEMALE,
            NEUTRAL and PLURAL.
    """
    gender = "NEUTRAL"
    head_index = attributes["head_index"]
    gender_data = external_data.GenderData.get_instance()

    if compute_number(attributes) == "PLURAL":
        gender = "PLURAL"
    elif attributes["type"] == "PRO":
        if attributes["citation_form"] == "he":
            gender = "MALE"
        elif attributes["citation_form"] == "she":
            gender = "FEMALE"
        elif attributes["citation_form"] == "it":
            gender = "NEUTRAL"
        elif attributes["citation_form"] in ["you", "we", "they"]:
            gender = "PLURAL"
    elif attributes["type"] == "NAM":
        if re.match(r"^mr(\.)?$", attributes["tokens"][0].lower()):
            gender = "MALE"
        elif re.match(r"^(miss|ms|mrs)(\.)?$",
                      attributes["tokens"][0].lower()):
            gender = "FEMALE"
        elif not re.match(r"(PERSON|NONE)", attributes["ner"][head_index]):
            gender = "NEUTRAL"
        elif gender_data.look_up(attributes):
            gender = gender_data.look_up(attributes)
    elif attributes["type"] == "NOM":
        if __wordnet_lookup_gender(" ".join(attributes["head"])):
            gender = __wordnet_lookup_gender(" ".join(attributes["head"]))
        elif gender_data.look_up(attributes):
            gender = gender_data.look_up(attributes)

    if gender == "NEUTRAL" and compute_semantic_class(attributes) == "PERSON":
        gender = "UNKNOWN"

    return gender


def compute_semantic_class(attributes):
    """ Compute the semantic class of a mention.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            values for "type", "head", "head_index" and, if the mention is a
            pronoun, "citation_form".

    Returns:
        str: the semantic class of the mention -- one of PERSON, OBJECT,
        NUMERIC and UNKNOWN.
    """
    semantic_class = "UNKNOWN"
    head_index = attributes["head_index"]

    if attributes["type"] == "PRO":
        if attributes["citation_form"] in ["i", "you", "he", "she", "we"]:
            semantic_class = "PERSON"
        elif attributes["citation_form"] == "they":
            semantic_class = "UNKNOWN"
        elif attributes["citation_form"] == "it":
            semantic_class = "OBJECT"
    elif attributes["type"] == "DEM":
        semantic_class = "OBJECT"
    elif attributes["ner"][head_index] != "NONE":
        ner_tag = attributes["ner"][head_index]
        if ner_tag == "PERSON":
            semantic_class = "PERSON"
        elif re.match("DATE|TIME|NUMBER|QUANTITY|MONEY|PERCENT", ner_tag):
            semantic_class = "NUMERIC"
        else:
            semantic_class = "OBJECT"
    # wordnet lookup
    elif (attributes["type"] == "NOM" and
            __wordnet_lookup_semantic_class(" ".join(attributes["head"]))):
        semantic_class = __wordnet_lookup_semantic_class(
            " ".join(attributes["head"]))

    return semantic_class


def __wordnet_lookup_semantic_class(head):
    synsets = wn.synsets(head)

    while synsets:
        lemma_name = synsets[0].lemma_names()[0]

        if lemma_name == "person":
            return "PERSON"
        elif lemma_name == "object":
            return "OBJECT"

        synsets = synsets[0].hypernyms()


def __wordnet_lookup_gender(head):
    synsets = wn.synsets(head)

    while synsets:
        lemma_name = synsets[0].lemma_names()[0]

        if lemma_name == "man" or lemma_name == "male":
            return "MALE"
        elif lemma_name == "woman" or lemma_name == "female":
            return "FEMALE"
        elif lemma_name == "person":
            return
        elif lemma_name == "entity":
            return "NEUTRAL"

        synsets = synsets[0].hypernyms()


def is_apposition(attributes):
    """ Compute whether the mention is an apposition, as in "Secretary of
    State Madeleine Albright" or "Barack Obama, the US president".

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            a value for "parse_tree".

    Returns:
        bool: Whether the mention is an apposition.
    """
    tree = attributes["parse_tree"]

    if tree.label() == "NP" and len(tree) > 1:
        if len(tree) == 2:
            return (tree[0].label() == "NP" and
                    tree[1].label() == "NP" and
                    __head_pos_starts_with(tree[1], "NNP"))
        elif len(tree) == 3:
            return (tree[0].label() == "NP" and
                    tree[1].label() == "," and
                    tree[2].label() == "NP" and
                    __any_child_head_starts_with(tree, "NNP") and
                    "DT" in set([child.pos()[0][1] for child in tree]))
        elif len(tree) == 4:
            return (tree[0].label() == "NP" and
                    tree[1].label() == "," and
                    tree[2].label() == "NP" and
                    tree[3].label() == "," and
                    __any_child_head_starts_with(tree, "NNP") and
                    "DT" in set([child.pos()[0][1] for child in tree]))


def __any_child_head_starts_with(tree, pos_tag):
    for child in tree:
        if __head_pos_starts_with(child, pos_tag):
            return True

    return False


def __head_pos_starts_with(tree, pos_tag):
    head_finder = head_finders.HeadFinder()
    return head_finder.get_head(tree).pos()[0][1].startswith(pos_tag)


def compute_head_information(attributes):
    """ Compute the head of the mention.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            values for "tokens", "parse_tree", "pos", "ner", "is_apposition"

    Returns:
        (list(str), Span, int): The head, the head span (in the document) and
        the starting index of the head (in the mention).
    """
    mention_subtree = attributes["parse_tree"]

    head_finder = head_finders.HeadFinder()
    head_index = 0
    head = [attributes["tokens"][0]]

    if len(mention_subtree.leaves()) == len(attributes["tokens"]):
        head_tree = head_finder.get_head(mention_subtree)
        head_index = get_head_index(head_tree, mention_subtree.pos())
        head = [head_tree[0]]

    in_mention_span = spans.Span(head_index, head_index)

    if attributes["pos"][head_index].startswith("NNP"):
        in_mention_span, head = \
            head_finders.HeadFinder.adjust_head_for_nam(
                attributes["tokens"],
                attributes["pos"],
                attributes["ner"][head_index])

    # proper name mention: head index last word of head
    # (e.g. "Obama" in "Barack Obama")
    head_index = in_mention_span.end

    # special handling for appositions
    if attributes["is_apposition"]:
        # "Secretary of State Madeleine Albright"
        # => take "Madeleine Albright" as head
        if len(mention_subtree) == 2:
            head_tree = mention_subtree[1]
            head = head_tree.leaves()
            in_mention_span = spans.Span(
                len(mention_subtree[0].leaves()),
                len(attributes["tokens"]) - 1)
            head_index = in_mention_span.end
        else:
            start = 0
            for child in mention_subtree:
                if __head_pos_starts_with(child, "NNP"):
                    end = min(
                        [start + len(child.leaves()),
                         len(attributes["tokens"])])
                    head_index = end - 1
                    in_mention_span, head = \
                        head_finders.HeadFinder.adjust_head_for_nam(
                            attributes["tokens"][start:end],
                            attributes["pos"][start:end],
                            attributes["ner"][head_index])
                    break
                start += len(child.leaves())

    return head, in_mention_span, head_index


def get_relevant_subtree(span, document):
    """ Get the fragment of the parse tree and the input span.

    Args:
        span (Span): A span in a document.
        document (CoNLLDocument): A document.

    Returns:
        nltk.ParentedTree: The fragment of the parse tree at the span in the
        document.
    """
    in_sentence_ids = document.in_sentence_ids[span.begin:span.end+1]
    in_sentence_span = spans.Span(in_sentence_ids[0], in_sentence_ids[-1])

    sentence_id, sentence_span = document.get_sentence_id_and_span(span)

    sentence_tree = document.parse[sentence_id]

    spanning_leaves = sentence_tree.treeposition_spanning_leaves(
        in_sentence_span.begin, in_sentence_span.end+1)
    mention_subtree = sentence_tree[spanning_leaves]

    if mention_subtree in sentence_tree.leaves():
        mention_subtree = sentence_tree[spanning_leaves[:-2]]

    return mention_subtree


def get_grammatical_function(attributes):
    """ Compute the grammatical function of a mention in its sentence.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            a value for "parse_tree".

    Returns:
        str: The grammatical function of the mention in its sentence, one of
        SUBJECT, OBJECT and OTHER.
    """
    tree = attributes["parse_tree"]
    parent = tree.parent()

    if parent is None:
        return "OTHER"
    else:
        parent_label = parent.label()

        if re.match(r"^(S|FRAG)", parent_label):
            return "SUBJECT"
        elif re.match(r"VP", parent_label):
            return "OBJECT"
        else:
            return "OTHER"


def get_head_index(head_with_pos, all_leaves):
    head_index = -1

    for i in range(0, len(all_leaves)):
        if head_with_pos[0] == all_leaves[i][0]:
            head_index = i

    return head_index


def get_type(attributes):
    """ Compute mention type.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            values for "pos", "ner" and "head_index".

    Returns:
        str: The mention type, one of NAM (proper name), NOM (common noun),
        PRO (pronoun), DEM (demonstrative pronoun) and VRB (verb).
    """
    pos = attributes["pos"][attributes["head_index"]]
    head_ner = attributes["ner"][attributes["head_index"]]

    if pos.startswith("NNP"):
        return "NAM"
    elif head_ner != "NONE":
        return "NAM"
    elif pos.startswith("PRP"):
        return "PRO"
    elif pos.startswith("DT"):
        return "DEM"
    elif pos.startswith("VB"):
        return "VRB"
    elif pos.startswith("NN"):
        return "NOM"
    else:
        return "NOM"


def get_fine_type(attributes):
    """ Compute fine-grained mention type.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            values for "type", "tokens" and "pos".

    Returns:
        str: The fine-grained mention type, one of

            - DEF (definite noun phrase),
            - INDEF (indefinite noun phrase),
            - PERS_NOM (personal pronoun, nominative case),
            - PERS_ACC (personal pronoun, accusative),
            - REFL (reflexive pronoun),
            - POSS (possessive pronoun),
            - POSS_ADJ (possessive adjective) or
            - None.
    """
    coarse_type = attributes["type"]
    start_token = attributes["tokens"][0]
    start_pos = attributes["pos"][0]

    if coarse_type == "NOM":
        if re.match("^(the|this|that|these|those|my|your|his|her|its|our|" +
                    "their)$", start_token.lower()):
            return "DEF"
        elif re.match("^NNP$", start_pos):  # also matches NNPS!
            return "DEF"
        else:
            return "INDEF"
    elif coarse_type == "PRO":
        if re.match("^(i|you|he|she|it|we|they)$", start_token.lower()):
            return "PERS_NOM"
        elif re.match("^(me|you|him|her|it|us|you|them)$",
                      start_token.lower()):
            return "PERS_ACC"
        elif re.match("^(myself|yourself|yourselves|himself|herself|itself|" +
                      "ourselves|themselves)$", start_token.lower()):
            return "REFL"
        elif start_pos == "PRP" and re.match("^(mine|yours|his|hers|its|" +
                                             "ours|theirs|)$",
                                             start_token.lower()):
            return "POSS"
        elif start_pos == "PRP$" and re.match("^(my|your|his|her|its|our|" +
                                              "their)$", start_token.lower()):
            return "POSS_ADJ"


def get_citation_form(attributes):
    """ Compute the citation form of a pronominal mention.

    Args:
        attributes (dict(str, object)): Attributes of the mention, must contain
            the key "tokens".

    Returns:
        str: The citation form of the pronoun, one of "i", "you", "he", "she",
        "it", "we", "they" and None.
    """
    pronoun = attributes["tokens"][0]

    pronoun = pronoun.lower()
    if re.match("^(he|him|himself|his)$",  pronoun):
        return "he"
    elif re.match("^(she|her|herself|hers|her)$",  pronoun):
        return "she"
    elif re.match("^(it|itself|its)$",  pronoun):
        return "it"
    elif re.match("^(they|them|themselves|theirs|their)$",  pronoun):
        return "they"
    elif re.match("^(i|me|myself|mine|my)$",  pronoun):
        return "i"
    elif re.match("^(you|yourself|yourselves|yours|your)$",  pronoun):
        return "you"
    elif re.match("^(we|us|ourselves|ours|our)$",  pronoun):
        return "we"



__author__ = 'smartschat'


class CorefMultigraphCreator:
    def __init__(self,
                 positive_features,
                 negative_features,
                 weighting_function,
                 relation_weights,
                 construct_when_negative=False):
        self.positive_features = positive_features
        self.negative_features = negative_features
        self.weighting_function = weighting_function
        self.relation_weights = relation_weights
        self.construct_when_negative = construct_when_negative

    def construct_graph_from_mentions(self, mentions):
        nodes = []
        edges = {}

        for i in range(0, len(mentions)):
            anaphor = mentions[i]

            nodes.append(anaphor)

            edges[anaphor] = self.construct_for_one_mention(mentions, i)

        return CorefMultigraph(nodes,
                               edges,
                               self.weighting_function,
                               self.relation_weights)

    def construct_for_one_mention(self, mentions, i):
        anaphor = mentions[i]

        edges = {}

        # do not include dummy mention
        for j in range(i-1, 0, -1):
            antecedent = mentions[j]
            if self.construct_when_negative:
                edges[antecedent] = self.get_edge_relations(anaphor, antecedent)
            else:
                if not self.has_negative(anaphor, antecedent):
                    edges[antecedent] = {
                        "negative_relations": [],
                        "positive_relations": self.get_positive_relations(
                            anaphor, antecedent)
                    }

        return edges

    def get_edge_relations(self, anaphor, antecedent):
        relations = {
            "negative_relations":
            self.get_negative_relations(anaphor, antecedent),
            "positive_relations":
            self.get_positive_relations(anaphor, antecedent)
        }

        return relations

    def has_negative(self, anaphor, antecedent):
        for r in self.negative_features:
            if r(anaphor, antecedent):
                return True

    def get_negative_relations(self, anaphor, antecedent):
        negative_relations = []

        for r in self.negative_features:
            if r(anaphor, antecedent):
                negative_relations.append(r)

        return negative_relations

    def get_positive_relations(self, anaphor, antecedent):
        positive_relations = []

        for r in self.positive_features:
            if r(anaphor, antecedent):
                positive_relations.append(r)

        return positive_relations


class CorefMultigraph:
    def __init__(self, nodes, edges, weighting_function, relation_weights):
        self.nodes = nodes
        self.edges = edges
        self.weighting_function = weighting_function
        self.relation_weights = relation_weights

    def get_weight(self, anaphor, antecedent):
        return self.weighting_function(
            anaphor,
            antecedent,
            self.edges[anaphor][antecedent],
            self.relation_weights)



import unittest

from cort.core.spans import Span


__author__ = 'smartschat'


class TestSpan(unittest.TestCase):
    def test_span(self):
        span = Span(0, 1)
        self.assertEqual(0, span.begin)
        self.assertEqual(1, span.end)

    def test_parse(self):
        self.assertEqual(Span(10, 12), Span.parse("(10, 12)"))
        self.assertEqual(Span(10, 12), Span.parse("(10,12)"))

if __name__ == '__main__':
    unittest.main()



""" Extract errors made by systems w.r.t. a reference corpus. """


from cort.analysis import data_structures


__author__ = 'smartschat'


class ErrorExtractor:
    """ Extract, manage and store recall and precision errors.

    Error extraction for recall errors works as follows:

    Go through each document. For each reference entity e in the document,
    construct an entity graph g_e for e and compute a partition of g_e by the
    system entity graphs. Then compute a spanning tree t_e of g_e and take
    every edge in t_e that does not appear in the partition as an error.

    For computing precision errors, switch the roles of reference and system
    entities.

    Attributes:
        reference_corpus (Corpus): The reference corpus with the gold
            information concerning the coreference relation.
        recall_spanning_tree_algorithm (function): A function mapping an
            entity graph and one its partitions to a list of mentions pairs,
            which represent a spanning tree of the entity graph. This
            function is used to compute recall errors.
        precision_spanning_tree_algorithm (function): Same as above, but for
            precision errors.
        errors (dict): A mapping of error descriptions to sets containing the
            respective errors.
    """
    def __init__(self,
                 reference_corpus,
                 recall_spanning_tree_algorithm,
                 precision_spanning_tree_algorithm,
                 ):
        """ Initialize the error analysis.

        Args:
            reference_corpus (Corpus): The reference corpus with the gold
                information concerning the coreference relation.
            recall_spanning_tree_algorithm (function): A function mapping an
                entity graph and one its partitions to a list of mentions pairs,
                which represent a spanning tree of the entity graph. This
                function is used to compute recall errors.
            precision_spanning_tree_algorithm (function): Same as above, but for
                precision errors.
        """

        self.reference_corpus = reference_corpus
        self.recall_spanning_tree_algorithm = recall_spanning_tree_algorithm
        self.precision_spanning_tree_algorithm = \
            precision_spanning_tree_algorithm
        self.errors = {}
        self.corpora = {}

    def add_system(self, system_corpus, which_mentions="annotated"):
        """ Add a system to the error analysis.

        Error extraction for recall errors works as follows:

        Go through each document. For each reference entity e in the document,
        construct an entity graph g_e for e and compute a partition of g_e by
        the system entity graphs. Then compute a spanning tree t_e of g_e and
        take every edge in t_e that does not appear in the partition as an
        error.

        For computing precision errors, switch the roles of reference and system
        entities.

        Also extracts all pairwise decisions (if available).

        Args:
            system_corpus (Corpus): A corpus obtained from system output.
            which_mentions (str): Either "annotated" or "extracted",
                defaults to "annotated". Specifies from which mentions in
                the system corpus coreference information should be
                obtained, either annotated mentions or system mentions.
        """
        if which_mentions not in ["annotated", "extracted"]:
            raise ValueError("which_mentions must be"
                             "either 'annotated' or 'extracted'.")

        recall_errors, precision_errors = self.__compute_errors(system_corpus,
                                                                which_mentions)

        self.errors[system_corpus.description] = {
            "recall_errors": {},
            "precision_errors": {},
            "decisions": {}
        }

        self.errors[system_corpus.description]["recall_errors"]["all"] = \
            recall_errors
        self.errors[
            system_corpus.description]["precision_errors"]["all"] = \
            precision_errors
        self.errors[
            system_corpus.description]["decisions"]["all"] = \
            system_corpus.get_antecedent_decisions()[
            system_corpus.description]["decisions"]["all"]

        self.corpora[system_corpus.description] = system_corpus

    def get_errors(self):
        """ Get errors for all systems managed by this ErrorAnalysis.

        The errors are stored via an ``StructuredCoreferenceAnalysis`
        which can be accessed like a dict.

        If a corpus with the description
        ``ranking``was added via ``self.add_system``,
        ``self.errors["ranking"]["recall_errors"]["all"]``is an ``EnhancedSet``
        containing all recall errors of the system. Errors of other systems
        and precision errors can be accessed analogously.

        Returns:
            StructuredCoreferenceAnalysis: A StructuredCoreferenceAnalysis
                containing the errors.
        """
        return data_structures.StructuredCoreferenceAnalysis(
            self.errors, corpora=self.corpora,
            reference=self.reference_corpus)

    def __compute_errors(self, system_corpus, which_mentions):
        gold_graphs = [data_structures.EntityGraph.from_mentions(
            doc.annotated_mentions, "annotated_set_id")
            for doc in self.reference_corpus.documents]

        if which_mentions == 'annotated':
            system_graphs = [data_structures.EntityGraph.from_mentions(
                doc.annotated_mentions, "annotated_set_id")
                for doc in system_corpus.documents]
        else:
            system_graphs = [data_structures.EntityGraph.from_mentions(
                doc.system_mentions, "set_id")
                for doc in system_corpus.documents]

        recall_errors = []
        precision_errors = []

        for doc_gold_graphs, doc_system_graphs in zip(gold_graphs,
                                                      system_graphs):
            recall_errors.extend(
                self.__compute_errors_for_doc(
                    doc_gold_graphs,
                    doc_system_graphs,
                    self.recall_spanning_tree_algorithm))
            precision_errors.extend(
                self.__compute_errors_for_doc(
                    doc_system_graphs,
                    doc_gold_graphs,
                    self.precision_spanning_tree_algorithm))

        return (data_structures.EnhancedSet(recall_errors),
                data_structures.EnhancedSet(precision_errors))

    @staticmethod
    def __compute_errors_for_doc(base_graphs,
                                 partitioning_graphs,
                                 spanning_tree_algorithm):
        errors = []

        for graph in base_graphs:
            errors.extend(
                ErrorExtractor.__compute_errors_for_graph(
                    graph, partitioning_graphs, spanning_tree_algorithm))

        return errors

    @staticmethod
    def __compute_errors_for_graph(graph,
                                   partitioning_graphs,
                                   spanning_tree_algorithm):
        partitioned_graph = graph.partition(partitioning_graphs)
        spanning_tree = spanning_tree_algorithm(graph, partitioned_graph)
        extra_pairs = [
            (anaphor, antecedent) for anaphor, antecedent in spanning_tree
            if anaphor not in partitioned_graph.edges or
            antecedent not in partitioned_graph.edges[anaphor]
        ]

        return [(anaphor, antecedent) for anaphor, antecedent in sorted(
            extra_pairs)]



""" Structures useful for error analysis. """


from collections import defaultdict


__author__ = 'smartschat'


class EntityGraph:
    """ Represent coreference relation between mentions in a graph.

    An entity graph consists of edges from anaphors to all coreferent mentions
    earlier in the text.

    More formally, an entity graph is a graph with the following properties:

        - the nodes are mentions
        - all edges point "backwards": if m appears later in the text than n,
          then there cannot exist an edge from n to m
        - there is an edge between to mentions if and only if they are
          coreferent

    Attributes:
        edges (dict(Mention, list(Mention))): A mapping from mentions to all
            mentions which have an incoming edge from that mention.

    """

    def __init__(self, edges):
        """ Initialize an entity graph from edges.

        Args:
            edges (dict(Mention, list(Mention))): A mapping from mentions to all
                mentions which have an incoming edge from that mention.
        """
        self.edges = edges

    def __eq__(self, other):
        """ Compare graphs for equality.

        Two graphs are equal if they have the same edges.

        Args:
            other (EntityGraph): An entity graph

        Returns:
            True if the graphs have the same edges, False otherwise.
        """
        if isinstance(other, self.__class__):
            return self.edges == other.edges
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(frozenset([(key, tuple(value))
                    for key, value in self.edges.items()]))

    def __repr__(self):
        return repr(self.edges)

    def __str__(self):
        return str(self.edges)

    @staticmethod
    def from_mentions(mentions, id_attribute):
        """ Construct a set of entity graphs from mentions.

        In particular, build a complete graph for every subset of mentions that
        have the same id_attribute.

        Args:
            mentions (list(Mention)): A list of mentions.
            id_attribute (str): The mention attribute which should be used to
                determine coreference. Possible values are "annotated_set_id"
                and "set_id".

        Returns:
            list(EntityGraph): A list of entity graphs. Two mentions are in one
            entity graph if and only if they have the same id_attribute.
        """
        id_to_mentions = defaultdict(list)

        for mention in mentions:
            if mention.attributes[id_attribute] is None:
                continue
            else:
                id_to_mentions[
                    mention.attributes[id_attribute]].append(mention)

        graphs = [EntityGraph.__create_complete(sorted(mention_list))
                      for mention_list in id_to_mentions.values()
                      if len(mention_list) > 1]

        return graphs

    @staticmethod
    def __create_complete(mentions):
        edges = {}
        for i in range(1, len(mentions)):
            edges[mentions[i]] = sorted(mentions[0:i], reverse=True)

        return EntityGraph(edges)

    def partition(self, entity_graphs):
        """ Partition the entity graph with respect to a set of entity graphs.

        The partitioned graph is a subgraph of this entity graph. An edge is
        retained if it is found in some graph in the supplied set of entity
        graphs.

        Args:
            entity_graphs (list(EntityGraph)): A list of entity graphs.

        Returns:
            EntityGraph: The partition of this graph with respect to the
            supplied graphs.
        """
        edges = {}
        for anaphor in self.edges:
            for antecedent in self.edges[anaphor]:
                if EntityGraph.__in_some_entity_graph(anaphor, antecedent,
                                                      entity_graphs):
                    if anaphor not in edges:
                        edges[anaphor] = list()
                    edges[anaphor].append(antecedent)

        return EntityGraph(edges)

    @staticmethod
    def __in_some_entity_graph(anaphor, antecedent, entity_graphs):
        for entity_graph in entity_graphs:
            if EntityGraph.__in_entity_graph(anaphor, antecedent, entity_graph):
                return True

    @staticmethod
    def __in_entity_graph(anaphor, antecedent, entity_graph):
        return anaphor in entity_graph.edges and \
            antecedent in entity_graph.edges[anaphor]

    def difference(self, entity_graph):
        """ Get all pairs of mention that are in this graph, but not in the
        supplied entity graph.

        Args:
            entity_graph (EntityGraph): An entity graph

        Returns:
            list((Mention, Mention): Pairs of mentions describing edges that
            can be found in this graph, but not in the supplied graph.
        """
        difference = []

        for anaphor in self.edges:
            for antecedent in self.edges[anaphor]:
                if (anaphor not in entity_graph.edges or
                        antecedent not in entity_graph.edges[anaphor]):
                    difference.append((anaphor, antecedent))

        return difference


class EnhancedSet:
    """ Manage, filter and categorize sets.

    For example, this class can manage coreference resolution errors or
    antecedent decisions.

    Attributes:
        data (set): A set.
    """

    def __init__(self, data=None):
        """ Initialize an enhanced set.

        Args:
            data: Any collection of data, defaults to None.
        """
        if data:
            self.data = set(data)
        else:
            self.data = set()

    def __iter__(self):
        return self.data.__iter__()

    def __contains__(self, datum):
        return datum in self.data

    def __len__(self):
        return len(self.data)

    def __eq__(self, other):
        if isinstance(other, EnhancedSet):
            return self.data == other.data
        else:
            return False

    def __hash__(self):
        return hash(self.data)

    def __repr__(self):
        return sorted(self.data).__repr__()

    def __str__(self):
        return sorted(self.data).__str__()

    def filter(self, function):
        """ Return a new EnhancedSet by filtering with respect to a function.

        All members ``x`` for which ``if function(x)`` holds are retained, the
        remaining are discarded. Typically, the function will be a boolean
        function, and therefore correspond to a predicate. For example, if the
        members are mention pairs corresponding to coreference resolution
        errors,
        ``lambda x: x[0].attributes["type"] == "NAM"``,
        will retain all errors with a proper name anaphor.

        Args:
            function: A function mapping members to values.

        Returns:
            EnhancedSet: An EnhancedSet filtered by the function.
        """
        filtered = set()

        for datum in self.data:
            if function(datum):
                filtered.add(datum)

        return EnhancedSet(filtered)

    def categorize(self, categorizer,
                   corpora=None,
                   reference=None):
        """ Categorize members with respect to a categorizer.

        Compute a category for each member, and then return a mapping of
        categories to EnhancedSets of members which have the respective
        category.

        For example, for coreference resolution errors, one typical
        categorizer is the mention type mapping of anaphor and
        antecedent, which is computed by the function
        ``lambda x: (x[0].attributes["type"], x[1].attributes["type"])``.

        Args:
            categorizer (function): A function mapping members to categories.

        Returns
            StructuredCoreferenceAnalysis: A StructuredCoreferenceAnalysis
                containing the categorized items. All items in one category
                can be accessed like in a dict, for example
                ``categorized[("NOM", "NOM")]``, if ``output`` is the
                StructuredCoreferenceAnalysis.
        """
        categorized = {}

        for datum in self.data:
            category = categorizer(datum)

            if category not in categorized:
                categorized[category] = EnhancedSet()

            categorized[category].data.add(datum)

        return StructuredCoreferenceAnalysis(categorized, corpora, reference)

    def intersection(self, other):
        """ Return the intersection of this EnhancedSet and another EnhancedSet.

        Args:
            other (EnhancedSet): An EnhancedSet.

        Returns
            EnhancedSet: All members which are in this EnhancedSet and in the
                other EnhancedSet.
        """
        return EnhancedSet(self.data.intersection(other.data))

    def difference(self, other):
        """ Return the difference of this EnhancedSet and another EnhancedSet.

        Args:
            other (EnhancedSet): An EnhancedSet.

        Returns
            EnhancedSet: All members which are in this EnhancedSet but not in
                the other EnhancedSet.
        """
        return EnhancedSet(self.data.difference(other.data))


class StructuredCoreferenceAnalysis:
    """ Manage EnhancedSets containing coreference information, for example
    errors or antecedent decisions.

    The included EnhancedSets can  be categorized and subcategorized to an
    arbitrary level. Access to these subcategorized sets works like in a dict,
    for example as in
    ``enhanced_mapping["pair"]["recall_errors"]["all"]``

    Attributes:
        mapping (dict(object, StructuredCoreferenceAnalysis)): A mapping of
            categories ``to StructuredCoreferenceAnalysis`` objects..
    """
    def __init__(self, input_mapping, corpora, reference):
        """ Initialize an StructuredCoreferenceAnalysis with a dict.

        Recursively transform the dict into a StructuredCoreferenceAnalysis.

        Args:
            input_mapping (dict): A nested dict. Leaves of the dict should be
                EnhancedSets.
            corpora (dict(str, Corpus)):  A mapping of corpus names to the
                respective Corpus objects. Stores information which system
                corpora are included in this StructuredCoreferenceAnalysis.
            reference (Corpus): The reference corpus for entries in this
                StructuredCoreferenceAnalysis.
        """
        self.corpora = corpora
        self.reference = reference
        mapping = {}

        for key, val in input_mapping.items():
            if isinstance(val, EnhancedSet):
                mapping[key] = val
            else:
                mapping[key] = StructuredCoreferenceAnalysis(
                    input_mapping[key], corpora, reference)

        self.mapping = mapping

    def __iter__(self):
        """ Iterate over all errors (recursively) contained in this
        StructuredCoreferenceAnalysis.
        """
        for key, val in self.mapping.items():
            if isinstance(val, StructuredCoreferenceAnalysis):
                for inner_val in val.__iter__():
                    yield inner_val
            else:
                for datum in val:
                    yield datum

    def __len__(self):
        """ Return the number of all errors contained in this
        StructuredCoreferenceAnalysis.

        Returns:
            int: The number of all errors contained in this
                StructuredCoreferenceAnalysis.
        """
        total_len = 0

        for key, val in self.mapping.items():
            total_len += len(val)

        return total_len

    def __eq__(self, other):
        if isinstance(other, StructuredCoreferenceAnalysis):
            return self.mapping == other.mapping
        else:
            return False

    def __hash__(self):
        return hash(self.mapping)

    def __repr__(self):
        return repr(self.mapping)

    def __str__(self):
        return str(self.mapping)

    def __getitem__(self, key):
        return self.mapping[key]

    def keys(self):
        return self.mapping.keys()

    def items(self):
        return self.mapping.items()

    def categorize(self, categorizer):
        """ Categorize EnhancedSets in this analysis with respect to a
        categorizer.

        Compute a category for each member in the EnhancedSets in this
        analysis. Add the respective categories at the highest nesting level.

        For example, for coreference resolution errors, one typical
        categorizer is the mention type mapping of anaphor and
        antecedent, which is computed by the function
        ``lambda x: (x[0].attributes["type"], x[1].attributes["type"])``.

        If the highest level before could be accessed as
        ``self["pair"]["recall_errors"]["all"]``,
        now the highest level can be accessed with
        ``self["pair"]["recall_errors"]["all"][("NOM", "NOM")]``,
        ``self["pair"]["recall_errors"]["all"][("PRO", "NAM")]``, ...

        Args:
            categorizer (function): A function mapping members to categories.

        Returns
            StructuredCoreferenceAnalysis: The analysis enriched with the new
                categories.
        """
        categorized_mapping = {}

        self._construct_helper(categorized_mapping, self.mapping, [],
                               categorizer, "categorize")

        return StructuredCoreferenceAnalysis(categorized_mapping,
                                             self.corpora,
                                             self.reference)

    def filter(self, function):
        """ Filter contained EnhancedSets with respect to a function.

        All members ``x`` in the EnhancedSets for which ``if function(x)``
        holds are retained, the remaining are discarded. Typically, the
        function will be a boolean  function, and therefore correspond to a
        predicate. For example, if the members are mention pairs
        corresponding to coreference resolution errors,
        ``lambda x: x[0].attributes["type"] == "NAM"``,
        will retain all errors with a proper name anaphor.

        Args:
            function: A function mapping members to values.

        Returns:
            StructuredCoreferenceAnalysis: A StructuredCoreferenceAnalysis
                where the EnhancedSets are filtered with respect to the
                function.
        """
        filtered_mapping = {}

        self._construct_helper(filtered_mapping, self.mapping, [],
                               function, "filter")

        return StructuredCoreferenceAnalysis(filtered_mapping,
                                             self.corpora,
                                             self.reference)

    def update(self, other):
        """ Update this StructuredCoreferenceAnalysis with another
        StructuredCoreferenceAnalysis.

        Entries from the other StructuredCoreferenceAnalysis have priority.

        Args:
            other (StructuredCoreferenceAnalysis): Another
                StructuredCoreferenceAnalysis.
        """
        self._construct_helper(self.mapping, other, [], lambda x: True,
                               "filter")

    def _construct_helper(self, constructed_mapping, mapping, position,
                          function, construct_type):
        curr_mapping = mapping
        for key in position:
            curr_mapping = curr_mapping[key]

        for key in curr_mapping.keys():
            val = curr_mapping[key]
            if isinstance(val, EnhancedSet):
                if construct_type == "categorize":
                    constructed_mapping[key] = val.categorize(function,
                                                              self.corpora,
                                                              self.reference)
                elif construct_type == "filter":
                    constructed_mapping[key] = val.filter(function)
            else:
                if key not in constructed_mapping:
                    new_corpora = self.corpora

                    if isinstance(mapping, StructuredCoreferenceAnalysis):
                        new_corpora.update(mapping.corpora)

                    constructed_mapping[key] = \
                        StructuredCoreferenceAnalysis({}, new_corpora,
                                                      self.reference)

                self._construct_helper(constructed_mapping[key].mapping,
                                       mapping,
                                       position + [key],
                                       function,
                                       construct_type)

    def visualize(self, corpus_name, error=None):
        """ Visualize errors contained in this StructuredCoreferenceAnalysis.

        In particular, visualize all recall and precision errors for the
        system ``corpus_name``. Errors can be categorized, but no nested
        categorization is supported for visualization. That is, visualize
        all errors that can be found under

        ``self[corpus_name]["recall_errors"]["all"]`` and
        ``self[corpus_name]["precision"]["all"]``,

        or, if the errors are categorized, under

        ``self[corpus_name]["recall_errors"]["all"][category]`` for each
        category ``category``.

        The errors are written to an HTML file, which is copied to
        ``temp/output/error_analysis.html``. This file is opened automatically.

        Args:
            corpus_name (str): The name (description attribute) of the corpus
                which shall be visualized.
            error ((Mention, Mention), optional): An error to highlight.
                Defaults to None.

        """
        from cort.analysis import visualization
        visualization.Visualizer(self, corpus_name, error).run()



import numpy
import pyximport
pyximport.install(setup_args={"include_dirs": numpy.get_include()})

import importlib
import inspect


__author__ = 'martscsn'


def import_from_path(name):
    splitted = name.split(".")
    package_name = ".".join(splitted[:-1])
    cls = splitted[-1]

    package = importlib.import_module(package_name)

    imported = getattr(package, cls)

    return imported


def get_features(filename):
    mention_features = []
    pairwise_features = []

    for line in open(filename).readlines():
        feature = import_from_path(line.strip())
        number_of_arguments = len(inspect.getargspec(feature)[0])

        if number_of_arguments == 1:
            mention_features.append(feature)
        elif number_of_arguments == 2:
            pairwise_features.append(feature)
        else:
            raise ValueError("Features must have one or two arguments, "
                             "feature " + line.strip() + " has " +
                             str(number_of_arguments) + " arguments.")

    return mention_features, pairwise_features



from cort.core.external_data import GenderData

__author__ = 'smartschat'

import unittest


class TestGenderData(unittest.TestCase):
    def setUp(self):
        self.gender_data = GenderData.get_instance()

    def test_look_up(self):
        self.assertEqual("NEUTRAL",
                         self.gender_data.look_up({"tokens": ["snafu"]}))

        self.assertEqual("FEMALE",
                         self.gender_data.look_up(
                             {"tokens": ["Barbara", "Bush"],
                              "head": ["Barbara", "Bush"]}))

        self.assertEqual("MALE",
                         self.gender_data.look_up({
                             "tokens": ["Footballer", "Zidane"],
                             "head": ["Zidane"]}))

if __name__ == '__main__':
    unittest.main()
